"""Contain the registry of generic indicators."""

from __future__ import annotations

from typing import TYPE_CHECKING

from icclim._core.constants import (
    GROUP_BY_METHOD,
    GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD,
    RESAMPLE_METHOD,
)
from icclim._core.generic.functions import (
    average,
    count_occurrences,
    deficit,
    difference_of_extremes,
    difference_of_means,
    excess,
    fraction_of_total,
    generic_sum,
    max_consecutive_occurrence,
    max_of_rolling_average,
    max_of_rolling_sum,
    maximum,
    mean_of_absolute_one_time_step_difference,
    mean_of_difference,
    min_of_rolling_average,
    min_of_rolling_sum,
    minimum,
    percentile,
    standard_deviation,
    sum_of_spell_lengths,
)
from icclim._core.generic.indicator import GenericIndicator
from icclim._core.model.registry import Registry
from icclim.exception import InvalidIcclimArgumentError

if TYPE_CHECKING:
    from icclim._core.climate_variable import ClimateVariable


def _check_single_var(
    climate_vars: list[ClimateVariable], indicator: GenericIndicator
) -> None:
    if len(climate_vars) > 1:
        msg = f"{indicator.name} can only be computed on a single variable."
        raise InvalidIcclimArgumentError(msg)


def _check_couple_of_vars(
    climate_vars: list[ClimateVariable],
    indicator: GenericIndicator,
) -> None:
    if len(climate_vars) != 2:
        msg = (
            f"{indicator.name} can only be computed on two variables sharing the same"
            f" unit (e.g. 2 temperatures). You can either provide a secondary variable"
            f" with `in_files` or `var_name`, or you can let icclim compute this"
            f" second variable as a subset of the first one using"
            f" `base_period_time_range`."
        )
        raise InvalidIcclimArgumentError(msg)


class GenericIndicatorRegistry(Registry[GenericIndicator]):
    """Registry of generic indicators."""

    def __init__(self) -> None:
        super().__init__()

    _item_class = GenericIndicator

    CountOccurrences = GenericIndicator(
        "count_occurrences",
        count_occurrences,
        definition="Count occurrences when threshold(s) are met"
        " (e.g. SU, Tx90p, RR1).",
    )
    MaxConsecutiveOccurrence = GenericIndicator(
        "max_consecutive_occurrence",
        max_consecutive_occurrence,
        definition="Count the maximum number of consecutive occurrences when"
        " threshold(s) are met (e.g. CDD, CSU, CWD).",
    )
    SumOfSpellLengths = GenericIndicator(
        "sum_of_spell_lengths",
        sum_of_spell_lengths,
        definition="Sum the lengths of each consecutive occurrence spell when"
        " threshold(s) are met. The minimum spell length is controlled by"
        " `min_spell_length` (e.g. WSDI, CSDI).",
    )
    Excess = GenericIndicator(
        "excess",
        excess,
        check_vars=_check_single_var,
        definition="Compute the excess over the given threshold. The excess is"
        " `sum(x[x>t] - t)` where x is the studied variable and t the threshold"
        " (e.g. GD4).",
    )
    Deficit = GenericIndicator(
        "deficit",
        deficit,
        check_vars=_check_single_var,
        definition="Compute the deficit below the given threshold. The deficit is"
        " `sum(t - x[x<t])` where x is the studied variable and t the threshold"
        " (e.g. HD17).",
    )
    FractionOfTotal = GenericIndicator(
        "fraction_of_total",
        fraction_of_total,
        check_vars=_check_single_var,
        definition="Compute the fraction of values meeting threshold(s) over the sum of"
        " every values (e.g. R75pTOT, R95pTOT).",
    )
    Maximum = GenericIndicator(
        "maximum",
        maximum,
        definition="Maximum of values that met threshold(s), if threshold(s) are given"
        " (e.g. Txx, Tnx).",
    )
    Minimum = GenericIndicator(
        "minimum",
        minimum,
        definition="Minimum of values that met threshold(s), if threshold(s) are given"
        " (e.g. Txn, Tnn).",
    )
    Average = GenericIndicator(
        "average",
        average,
        definition="Average of values that met threshold(s), if threshold(s) are given"
        " (e.g. Tx, Tn).",
    )
    Sum = GenericIndicator(
        "sum",
        generic_sum,
        definition="Sum of values that met threshold(s), if threshold(s) are given"
        " (e.g. PRCPTOT, RR).",
    )
    StandardDeviation = GenericIndicator(
        "standard_deviation",
        standard_deviation,
        definition="Standard deviation of values that met threshold(s),"
        " if threshold(s) are given.",
    )
    MaxOfRollingSum = GenericIndicator(
        "max_of_rolling_sum",
        max_of_rolling_sum,
        check_vars=_check_single_var,
        definition="Maximum of rolling sum over time dimension"
        " (e.g. RX5DAY: maximum 5 days window of precipitation accumulation).",
    )
    MinOfRollingSum = GenericIndicator(
        "min_of_rolling_sum",
        min_of_rolling_sum,
        check_vars=_check_single_var,
        definition="Minimum of rolling sum over time dimension.",
    )
    MaxOfRollingAverage = GenericIndicator(
        "max_of_rolling_average",
        max_of_rolling_average,
        check_vars=_check_single_var,
        definition="Maximum of rolling average over time dimension.",
    )
    MinOfRollingAverage = GenericIndicator(
        "min_of_rolling_average",
        min_of_rolling_average,
        check_vars=_check_single_var,
        definition="Minimum of rolling average over time dimension.",
    )
    MeanOfDifference = GenericIndicator(
        "mean_of_difference",
        mean_of_difference,
        check_vars=_check_couple_of_vars,
        definition="Average of the difference between two variables"
        ", or one variable and it's reference period values"
        " (e.g. DTR: `mean(tasmax - tasmin)`).",
        qualifiers=["compute_diff"],
    )
    DifferenceOfExtremes = GenericIndicator(
        "difference_of_extremes",
        difference_of_extremes,
        check_vars=_check_couple_of_vars,
        definition="Difference of extremes between two variables"
        ", or one variable and it's reference period values."
        " The extremes are always `maximum` for the first variable and"
        " `minimum` for the second variable"
        " (e.g. ETR: `max(tasmax) - min(tasmin)`).",
        qualifiers=["compute_diff"],
    )
    MeanOfAbsoluteOneTimeStepDifference = GenericIndicator(
        "mean_of_absolute_one_time_step_difference",
        mean_of_absolute_one_time_step_difference,
        check_vars=_check_couple_of_vars,
        definition="Average of the absolute one time step by one time step difference"
        " between two variables,"
        " or one variable and it's reference period values"
        " (e.g. vDTR:"
        " `mean((tasmax[i] - tasmin[i]) - (tasmax[i-1] - tasmin[i-1])` ;"
        " where i is the day of measure).",
        qualifiers=["compute_diff"],
    )
    DifferenceOfMeans = GenericIndicator(
        "difference_of_means",
        difference_of_means,
        check_vars=_check_couple_of_vars,
        sampling_methods=[
            RESAMPLE_METHOD,
            GROUP_BY_METHOD,
            GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD,
        ],
        definition="Difference of the average between two variables"
        ", or one variable and it's reference period values"
        " (e.g. anomaly: `mean(tasmax) - mean(tasmax_ref]))`.",
        qualifiers=["compute_diff"],
    )
    Percentile = GenericIndicator(
        "percentile",
        percentile,
        check_vars=_check_single_var,
        definition="Percentile of a variable.",
    )



"""
Implementations of the generic indices computation methods.

These functions are not meant to be called directly, they are used by the
`GenericIndicatorRegistry` class to create the generic indices.
Each function is a reducer that takes a list of `ClimateVariable` instances and returns
a `DataArray` instance.
The `ClimateVariable` instances are used to extract the data and the thresholds needed
to compute the generic index.
The `DataArray` instance is the result of the computation of the generic index.

.. note::

   You can call the respective generic index from icclim module, for example:
   `icclim.count_occurrences(...)`.
"""

from __future__ import annotations

import operator
from collections.abc import Callable
from functools import partial
from typing import TYPE_CHECKING
from warnings import warn

import xarray as xr
from numpy import abs as np_abs
from numpy import diff as np_diff
from numpy import median as np_median
from pandas import Timedelta, date_range, infer_freq, to_timedelta
from xarray import DataArray
from xarray.computation.rolling import DataArrayRolling
from xarray.core.resample import DataArrayResample
from xclim.core.calendar import build_climatology_bounds
from xclim.core.units import (
    convert_units_to,
    rate2amount,
    str2pint,
    to_agg_units,
)
from xclim.indices import run_length

from icclim._core.climate_variable import must_run_bootstrap
from icclim._core.constants import (
    GROUP_BY_METHOD,
    GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD,
    PART_OF_A_WHOLE_UNIT,
    REFERENCE_PERIOD_ID,
    RESAMPLE_METHOD,
    UNITS_KEY,
)
from icclim._core.input_parsing import PercentileDataArray
from icclim._core.model.cf_calendar import CfCalendarRegistry
from icclim._core.model.operator import OperatorRegistry
from icclim.exception import InvalidIcclimArgumentError
from icclim.frequency import RUN_INDEXER, Frequency, FrequencyRegistry

if TYPE_CHECKING:
    from datetime import timedelta

    import numpy as np
    from pint import Quantity
    from xarray.core.groupby import DataArrayGroupBy

    from icclim._core.climate_variable import ClimateVariable
    from icclim._core.generic.threshold.percentile import PercentileThreshold
    from icclim._core.model.logical_link import LogicalLink
    from icclim._core.model.threshold import Threshold


def count_occurrences(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    logical_link: LogicalLink,
    date_event: bool,
    to_percent: bool,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Count the occurrences of exceedances of the threshold(s).

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        The list of climate variables containing the data and the threshold(s).
    resample_freq : Frequency
        The time frequency of the output.
    logical_link : LogicalLink
        The logical link to apply to the exceedances if multiple thresholds are
        provided.
    date_event : bool
        Whether to return the date of the event.
    to_percent : bool
        Whether to return the result in percent.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The count of occurrences of exceedances.
    """
    if date_event:
        reducer_op = _count_occurrences_with_date
    else:
        reducer_op = partial(DataArrayResample.sum, dim="time")
    merged_exceedances = _compute_exceedances(
        climate_vars,
        resample_freq.pandas_freq,
        logical_link,
    )
    result = reducer_op(merged_exceedances.resample(time=resample_freq.pandas_freq))
    if to_percent:
        result = _to_percent(result, resample_freq)
        result.attrs[UNITS_KEY] = "%"
        return result
    freq = check_freq(climate_vars[0].studied_data, dim="time")
    return to_agg_units(result, climate_vars[0].studied_data, "count", deffreq=freq)


def max_consecutive_occurrence(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    logical_link: LogicalLink,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Calculate the maximum number of consecutive occurrences of exceedances for a given set of climate variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        The list of climate variables containing the data and the threshold(s).
    resample_freq : Frequency
        The time frequency of the output.
    logical_link : LogicalLink
        The logical link to apply when merging the exceedances.
    date_event : bool
        Whether to include the dates of the exceedances in the result.
    source_freq_delta : timedelta
        The time difference between consecutive data points in the source data.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The maximum number of consecutive occurrences of exceedances.
    """  # noqa: E501
    merged_exceedances = _compute_exceedances(
        climate_vars,
        resample_freq.pandas_freq,
        logical_link,
    )
    rle = run_length.rle(merged_exceedances, dim="time", index="first")
    resampled = rle.resample(time=resample_freq.pandas_freq)
    if date_event:
        result = _consecutive_occurrences_with_dates(resampled, source_freq_delta)
    else:
        result = resampled.max(dim="time")
    freq = check_freq(climate_vars[0].studied_data, dim="time")
    return to_agg_units(result, climate_vars[0].studied_data, "count", deffreq=freq)


def sum_of_spell_lengths(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    logical_link: LogicalLink,
    min_spell_length: int,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Calculate the sum of the lengths of all spells in the data.

    This function calculates the sum of the lengths of all spells in the data,
    where a spell is defined as a consecutive occurrence of exceedances.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        The list of climate variables containing the data and the threshold(s).
    resample_freq : Frequency
        The time frequency of the output.
    logical_link : LogicalLink
        The logical link to apply when merging the exceedances.
    min_spell_length : int
        The minimum length of a spell to consider.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The sum of the lengths of all spells in the data.
    """
    merged_exceedances = _compute_exceedances(
        climate_vars,
        resample_freq.pandas_freq,
        logical_link,
    )
    rle = run_length.rle(merged_exceedances, dim="time", index="first")
    cropped_rle = rle.where(rle >= min_spell_length, other=0)
    result = cropped_rle.resample(time=resample_freq.pandas_freq).max(dim="time")
    freq = check_freq(climate_vars[0].studied_data, dim="time")
    return to_agg_units(result, climate_vars[0].studied_data, "count", deffreq=freq)


def excess(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Compute the excess of a climate variable above a threshold using the 'reach' operator.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables. Only the first variable is used.
    resample_freq : Frequency
        The time frequency of the output.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        DataArray containing the computed excess values.

    Raises
    ------
    InvalidIcclimArgumentError
        If the threshold operator is not 'reach'.

    Notes
    -----
    The excess is computed by subtracting the threshold from the climate variable data.
    Only the values above the threshold are considered, and negative values are set to
    zero.
    The resulting excess values are then summed over the specified resample frequency.
    """  # noqa: E501
    study, threshold = _get_thresholded_var(climate_vars)
    if threshold.operator != OperatorRegistry.REACH:
        msg = "Excess can only be computed with 'reach' operator."
        raise InvalidIcclimArgumentError(msg)
    excesses = threshold.compute(study, override_op=operator.sub)
    res = (
        (excesses).clip(min=0).resample(time=resample_freq.pandas_freq).sum(dim="time")
    )
    res = res.assign_attrs(units=f"delta_{res.attrs['units']}")
    freq = check_freq(study, dim="time")
    return to_agg_units(res, study, "integral", deffreq=freq)


def deficit(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Compute the deficit of a climate variable below a threshold using the 'reach' operator.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables. Only the first variable is used.
    resample_freq : Frequency
        The time frequency of the output.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        DataArray containing the computed deficit values.

    Notes
    -----
    The deficit is computed by subtracting the climate variable data from the threshold.
    Only the values below the threshold are considered, and negative values are set to
    zero.
    The resulting deficit values are then summed over the specified resample frequency.
    """  # noqa: E501
    study, threshold = get_single_var(climate_vars)
    deficit = threshold.compute(study, override_op=lambda da, th: th - da)
    res = deficit.clip(min=0).resample(time=resample_freq.pandas_freq).sum(dim="time")
    res = res.assign_attrs(units=f"delta_{res.attrs['units']}")
    freq = check_freq(study, dim="time")
    return to_agg_units(res, study, "integral", deffreq=freq)


def fraction_of_total(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    to_percent: bool,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Calculate the fraction of total for a given set of climate variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        The list of climate variables containing the data and the threshold(s).
        Only one variable is expected in the list.
    resample_freq : Frequency
        The resampling frequency.
    to_percent : bool
        Flag indicating whether to convert the result to percentage.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The fraction of total as a DataArray.

    Notes
    -----
    This function calculates the fraction of total for a given set of climate variables.
    The fraction of total is calculated by dividing the sum of values exceeding a
    threshold by the total sum of values.

    If the `to_percent` flag is set to True, the result will be multiplied by 100 and
    the units will be set to "%". Otherwise, the units will be set to the value of
    PART_OF_A_WHOLE_UNIT, which is 1.
    """
    study, threshold = get_single_var(climate_vars)
    if threshold.threshold_min_value is not None:
        min_val = threshold.threshold_min_value
        min_val = convert_units_to(min_val, study, context="hydro")
        total = (
            study.where(threshold.operator(study, min_val))
            .resample(time=resample_freq.pandas_freq)
            .sum(dim="time")
        )
    else:
        total = study.resample(time=resample_freq.pandas_freq).sum(dim="time")
    exceedance = _compute_exceedance(
        study=study,
        threshold=threshold,
        freq=resample_freq.pandas_freq,
        bootstrap=must_run_bootstrap(study, threshold),
    ).squeeze()
    over = (
        study.where(exceedance, 0)
        .resample(time=resample_freq.pandas_freq)
        .sum(dim="time")
    )
    res = over / total
    if to_percent:
        res = res * 100
        res.attrs[UNITS_KEY] = "%"
    else:
        res.attrs[UNITS_KEY] = PART_OF_A_WHOLE_UNIT
    return res


def maximum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    date_event: bool,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Calculate the maximum value of the given climate variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to calculate the maximum value for.
    resample_freq : Frequency
        The frequency at which to resample the data.
    date_event : bool
        Flag indicating whether the output should include the date of the events.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The maximum value of the climate variables.
    """
    return _run_simple_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        reducer_op=DataArrayResample.max,
        date_event=date_event,
    )


def minimum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    date_event: bool,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Calculate the minimum value of the given climate variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to calculate the minimum value for.
    resample_freq : Frequency
        The frequency at which to resample the data.
    date_event : bool
        Flag indicating whether the output should include the date of the events.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The minimum value of the climate variables.
    """
    return _run_simple_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        reducer_op=DataArrayResample.min,
        date_event=date_event,
    )


def average(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Compute the average of the given climate variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to compute the average for.
    resample_freq : Frequency
        The frequency at which to resample the data.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The computed average as a DataArray.

    """
    return _run_simple_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        reducer_op=DataArrayResample.mean,
        date_event=False,
    )


def generic_sum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Compute the sum of the given climate variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to compute the sum for.
    resample_freq : Frequency
        The frequency at which to resample the data.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The computed sum as a DataArray.
    """
    return _run_simple_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        reducer_op=DataArrayResample.sum,
        date_event=False,
        must_convert_rate=True,
    )


def standard_deviation(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Compute the standard deviation of the given climate variables.

    This function calculates the standard deviation of the provided climate variables.
    The standard deviation is a measure of the amount of variation or dispersion in the
    data.
    It quantifies the amount of variation or spread in the values of the climate
    variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to compute the standard deviation for.
    resample_freq : Frequency
        The frequency at which to resample the data.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The computed standard deviation as a DataArray.
    """
    return _run_simple_reducer(
        climate_vars,
        resample_freq,
        DataArrayResample.std,
        date_event=False,
    )


def max_of_rolling_sum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Compute the maximum value of the rolling sum of the given climate variables.

    The rolling sum is the sum of values over a specified rolling window width.
    The maximum value is the highest value obtained from the rolling sum.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to compute the maximum value of the rolling sum for.
    resample_freq : Frequency
        The frequency at which to resample the data.
    rolling_window_width : int
        The width of the rolling window, i.e., the number of values to include in each
        rolling sum.
    date_event : bool
        A flag indicating whether the date of the events should be included in the
        output.
    source_freq_delta : timedelta
        The time difference between consecutive data points in the source data.
        For daily data this is 1 day, for monthly data this is 1 month, etc.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The computed maximum value of the rolling sum as a DataArray.
    """
    return _run_rolling_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        rolling_window_width=rolling_window_width,
        rolling_op=DataArrayRolling.sum,
        resampled_op=DataArrayResample.max,
        date_event=date_event,
        source_freq_delta=source_freq_delta,
    )


def min_of_rolling_sum(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Compute the minimum value of the rolling sum of the given climate variables.

    The rolling sum is the sum of values over a specified rolling window width.
    The minimum value is the lowest value obtained from the rolling sum.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to compute the minimum value of the rolling sum for.
    resample_freq : Frequency
        The frequency at which to resample the data.
    rolling_window_width : int
        The width of the rolling window, i.e., the number of values to include in each
        rolling sum.
    date_event : bool
        A flag indicating whether the date of the events should be included in the
        output.
    source_freq_delta : timedelta
        The time difference between consecutive data points in the source data.
        For daily data this is 1 day, for monthly data this is 1 month, etc.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The computed minimum value of the rolling sum as a DataArray.
    """
    return _run_rolling_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        rolling_window_width=rolling_window_width,
        rolling_op=DataArrayRolling.sum,
        resampled_op=DataArrayResample.min,
        date_event=date_event,
        source_freq_delta=source_freq_delta,
    )


def min_of_rolling_average(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Compute the minimum value of the rolling average of the given climate variables.

    The rolling average is the average of values over a specified rolling window width.
    The minimum value is the lowest value obtained from the rolling average.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to compute the minimum value of the rolling average
        for.
    resample_freq : Frequency
        The frequency at which to resample the data.
    rolling_window_width : int
        The width of the rolling window, i.e., the number of values to include in each
        rolling average.
    date_event : bool
        A flag indicating whether the date of the events should be included in the
        output.
    source_freq_delta : timedelta
        The time difference between consecutive data points in the source data.
        For daily data this is 1 day, for monthly data this is 1 month, etc.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The computed minimum value of the rolling average as a DataArray.
    """
    return _run_rolling_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        rolling_window_width=rolling_window_width,
        rolling_op=DataArrayRolling.mean,
        resampled_op=DataArrayResample.min,
        date_event=date_event,
        source_freq_delta=source_freq_delta,
    )


def max_of_rolling_average(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    date_event: bool,
    source_freq_delta: timedelta,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Compute the minimum value of the rolling average of the given climate variables.

    The rolling average is the average of values over a specified rolling window width.
    The minimum value is the lowest value obtained from the rolling average.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to compute the minimum value of the rolling average
        for.
    resample_freq : Frequency
        The frequency at which to resample the data.
    rolling_window_width : int
        The width of the rolling window, i.e., the number of values to include in each
        rolling average.
    date_event : bool
        A flag indicating whether the date of the events should be included in the
        output.
    source_freq_delta : timedelta
        The time difference between consecutive data points in the source data.
        For daily data this is 1 day, for monthly data this is 1 month, etc.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The computed minimum value of the rolling average as a DataArray.
    """
    return _run_rolling_reducer(
        climate_vars=climate_vars,
        resample_freq=resample_freq,
        rolling_window_width=rolling_window_width,
        rolling_op=DataArrayRolling.mean,
        resampled_op=DataArrayResample.max,
        date_event=date_event,
        source_freq_delta=source_freq_delta,
    )


def mean_of_difference(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Calculate the mean of the difference between two climate variables.

    This function calculates the mean of the difference between two climate variables
    for each time step, and then resamples the resulting data based on the specified
    frequency.
    The resulting data array will have the same units as the study variable.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        The two climate variables necessary to compute the indicator.
    resample_freq : Frequency
        Resampling frequency of the output.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The mean of the difference as a xarray.DataArray.

    Notes
    -----
    This is a generification of ECAD's DTR climate index.
    """
    study, ref = get_couple_of_var(climate_vars, "mean_of_difference")
    mean_of_diff = (study - ref).resample(time=resample_freq.pandas_freq).mean()
    mean_of_diff.attrs["units"] = study.attrs["units"]
    return mean_of_diff


def difference_of_extremes(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Calculate the difference of extremes between two climate variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        A list of climate variables.
    resample_freq : Frequency
        The frequency at which to resample the data.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The difference of extremes between the two climate variables.

    Notes
    -----
    This function calculates the difference of extremes between two climate variables.
    It first resamples the study variable to the specified frequency and take the
    maximum per resampled chunk.
    Then it resamples the reference variable to the same frequency and take the minimum
    per resampled chunk.
    Finally, for each chunk, it calculates the differences of theses maximum and
    minimum values.
    This is a generification of ECAD's ETR climate index.
    """
    study, ref = get_couple_of_var(climate_vars, "difference_of_extremes")
    max_study = study.resample(time=resample_freq.pandas_freq).max()
    min_ref = ref.resample(time=resample_freq.pandas_freq).min()
    diff_of_extremes = max_study - min_ref
    diff_of_extremes.attrs["units"] = study.attrs["units"]
    return diff_of_extremes


def mean_of_absolute_one_time_step_difference(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Mean of the absolute one-time-step difference between two climate variables.

    This function calculates the mean of the absolute difference between two climate
    variables
    for each time step, and then resamples the resulting data based on the specified
    frequency.
    The resulting data array will have the same units as the study variable.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        The two climate variables necessary to compute the indicator.
    resample_freq : Frequency
        Resampling frequency of the output.
    **kwargs : dict
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The mean of the absolute one-time-step difference as a xarray.DataArray.

    Notes
    -----
    This is a generification of ECAD's vDTR climate index.
    """
    study, ref = get_couple_of_var(
        climate_vars,
        "mean_of_absolute_one_time_step_difference",
    )
    one_time_step_diff = (study - ref).diff(dim="time")
    res = abs(one_time_step_diff).resample(time=resample_freq.pandas_freq).mean()
    res.attrs["units"] = study.attrs["units"]
    return res


def difference_of_means(
    climate_vars: list[ClimateVariable],
    to_percent: bool,
    resample_freq: Frequency,
    sampling_method: str,
    is_compared_to_reference: bool,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Calculate the difference of means between two climate variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        A studied climate variable and a reference climate variable.
    to_percent : bool
        If True, the result will be converted to percentage.
    resample_freq : Frequency
        Resampling frequency of the output.
    sampling_method : str
        The method used for resampling. It can be either 'group_by', 'resample', or
        'group_by_ref_and_resample_study'.
        'group_by' will group the data by the specified frequency, for example every
        data of every January together.
        'resample' will resample the data to the specified frequency, for example every
        days of each month independently together.
        'group_by_ref_and_resample_study' will group the reference data by the specified
        frequency and resample the study data to the same frequency.
        This last method allows for example to compare each January, independently, of
        the study period to every January of the reference period.
        This is typically used to compare the each month of the studied period
        to a normal (the reference) of many aggregated years.
    is_compared_to_reference : bool
        If True, check if the sampling method is 'resample' and raise an error if it is.
        It does not make sense to resample the reference variable if it is already a
        subsample of the studied variable.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The difference of means between the two climate variables.

    Notes
    -----
    This is a generification of the anomaly climate index.
    """
    study = climate_vars[0].studied_data
    ref = climate_vars[1].studied_data
    study = convert_units_to(study, ref, context="hydro")
    return _reduce_and_diff(
        study,
        ref,
        to_percent,
        resample_freq,
        sampling_method,
        is_compared_to_reference,
        reducer=lambda x: x.mean(dim="time"),
    )


def percentile(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    """
    Calculate the percentile of the given climate variable.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        A single climate variable within a list.
    resample_freq : Frequency
        Resampling frequency of the output.
    **kwargs
        Ignored keyword arguments (for compatibility).

    Returns
    -------
    DataArray
        The calculated percentile as a DataArray.

    Notes
    -----
    This function calculates the percentile of the given climate variables
    by resampling the data based on the provided frequency and then
    calculating the corresponding quantile using the specified interpolation method.

    The resulting DataArray contains the percentiles as the 'percentiles'
    coordinate variable.
    """
    study, threshold = get_single_var(climate_vars)
    threshold: PercentileThreshold
    quantile = threshold.initial_value[0] * 0.01
    method = threshold.interpolation.name
    result = study.resample(time=resample_freq.pandas_freq).quantile(
        quantile, method=method
    )
    result.coords["quantile"] = result.coords["quantile"] * 100
    result = result.rename(quantile="percentiles")
    return PercentileDataArray.from_da(
        source=result,
        climatology_bounds=build_climatology_bounds(study),
    )


def _reduce_and_diff(
    study: DataArray,
    ref: DataArray,
    to_percent: bool,
    resample_freq: Frequency,
    sampling_method: str,
    is_compared_to_reference: bool,
    reducer: Callable[[DataArrayResample | DataArray | DataArrayGroupBy], DataArray],
    use_reduce_unit: bool = False,
    **kwargs,  # noqa: ARG001
) -> DataArray:
    if is_compared_to_reference and sampling_method == RESAMPLE_METHOD:
        msg = (
            "It does not make sense to resample the reference variable if it is"
            " already a subsample of the studied variable. Try setting"
            f" `sampling_method='{GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD}'`"
            f" instead."
        )
        raise InvalidIcclimArgumentError(msg)
    if sampling_method == GROUP_BY_METHOD:
        if resample_freq.group_by_key == RUN_INDEXER:
            reduced_study = reducer(study)
            reduced_ref = reducer(ref)
        else:
            reduced_study = reducer(study.groupby(resample_freq.group_by_key))
            reduced_ref = reducer(ref.groupby(resample_freq.group_by_key))
    elif sampling_method == RESAMPLE_METHOD:
        reduced_study = reducer(study.resample(time=resample_freq.pandas_freq))
        reduced_ref = reducer(ref.resample(time=resample_freq.pandas_freq))
    elif sampling_method == GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD:
        if (
            resample_freq.group_by_key == RUN_INDEXER
            or resample_freq == FrequencyRegistry.YEAR
        ):
            reduced_study = reducer(study.resample(time=resample_freq.pandas_freq))
            # data is already filtered with only the indexed values.
            # Thus there is only one "group".
            reduced_ref = reducer(ref)
        else:
            return _reduce_and_diff_of_resampled_x_by_groupedby_y(
                resample_freq, to_percent, study, ref, reducer=reducer
            )
    else:
        msg = f"Unknown sampling_method: '{sampling_method}'."
        raise NotImplementedError(msg)
    diff = reduced_study - reduced_ref
    if to_percent:
        diff = diff / reduced_ref * 100
        diff.attrs[UNITS_KEY] = "%"
    elif use_reduce_unit:
        diff.attrs[UNITS_KEY] = reduced_study.attrs[UNITS_KEY]
    else:
        diff.attrs[UNITS_KEY] = study.attrs[UNITS_KEY]
    return diff


def _reduce_and_diff_of_resampled_x_by_groupedby_y(
    resample_freq: Frequency,
    to_percent: bool,
    study: DataArray,
    ref: DataArray,
    reducer: Callable[
        [DataArrayResample | DataArrayGroupBy | DataArray | ClimateVariable], DataArray
    ],
) -> DataArray:
    mean_ref = reducer(ref.groupby(resample_freq.group_by_key))
    acc = []
    if resample_freq == FrequencyRegistry.MONTH:
        key = "month"
        dt_selector = lambda x: x.time.dt.month  # noqa: E731
    elif resample_freq == FrequencyRegistry.DAY:
        key = "dayofyear"
        dt_selector = lambda x: x.time.dt.dayofyear  # noqa: E731
    else:
        msg = (
            f"Can't use {GROUP_BY_REF_AND_RESAMPLE_STUDY_METHOD}"
            f" with the frequency {resample_freq.long_name}."
        )
        raise NotImplementedError(msg)
    for label, sample in study.resample(time=resample_freq.pandas_freq):
        sample_mean = reducer(sample)
        ref_group_mean = mean_ref.sel({key: dt_selector(sample).to_numpy()[0]})
        sample_diff_of_means = sample_mean - ref_group_mean
        if to_percent:
            sample_diff_of_means = sample_diff_of_means / ref_group_mean * 100
        del sample_diff_of_means[key]
        sample_diff_of_means = sample_diff_of_means.expand_dims(time=[label])
        acc.append(sample_diff_of_means)
    diff_of_means = xr.concat(acc, dim="time")
    if to_percent:
        diff_of_means.attrs["units"] = "%"
    else:
        diff_of_means.attrs["units"] = study.attrs["units"]
    return diff_of_means


def _compute_exceedance(
    study: DataArray,
    threshold: Threshold,
    freq: str,  # used by @percentile_bootstrap (don't rename, it breaks bootstrap)
    bootstrap: bool,  # used by @percentile_bootstrap
) -> DataArray:
    exceedances = threshold.compute(study, freq=freq, bootstrap=bootstrap)
    if bootstrap:
        exceedances.attrs[REFERENCE_PERIOD_ID] = threshold.value.attrs[
            "climatology_bounds"
        ]
    return exceedances


def get_couple_of_var(
    climate_vars: list[ClimateVariable],
    indicator: str,
) -> tuple[DataArray, DataArray]:
    """
    Get exactly two climate variables to compute a climate indicator.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        A list of climate variables.
    indicator : str
        The name of the indicator to be computed.

    Returns
    -------
    tuple[DataArray, DataArray]
        A tuple containing two DataArray objects representing the study variable and
        the reference variable.

    Raises
    ------
    InvalidIcclimArgumentError
        If the number of climate variables is not equal to 2.
        If any of the two variable has a threshold.

    Notes
    -----
    This function is used to extract a couple of climate variables needed for computing
    an indicator.
    The function checks the number of climate variables and raises an error
    if it is not equal to 2 or if thresholds are present.
    """
    if len(climate_vars) != 2:
        msg = (
            f"{indicator} needs two variables **or** one variable and a "
            f"`base_period_time_range` period to extract a reference variable."
        )
        raise InvalidIcclimArgumentError(msg)
    if climate_vars[0].threshold or climate_vars[1].threshold:
        msg = f"{indicator} cannot be computed with thresholds."
        raise InvalidIcclimArgumentError(msg)
    study = climate_vars[0].studied_data
    ref = climate_vars[1].studied_data
    study = convert_units_to(study, ref, context="hydro")
    return study, ref


def _run_rolling_reducer(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    rolling_window_width: int,
    rolling_op: Callable[[DataArrayRolling], DataArray],  # sum | mean
    resampled_op: Callable[[DataArrayResample], DataArray],  # max | min
    date_event: bool,
    source_freq_delta: timedelta,
) -> DataArray:
    study, threshold = get_single_var(climate_vars)
    if threshold:
        exceedance = _compute_exceedance(
            study=study,
            freq=resample_freq.pandas_freq,
            threshold=threshold,
            bootstrap=must_run_bootstrap(study, threshold),
        ).squeeze()
        study = study.where(exceedance)
    study = rolling_op(study.rolling(time=rolling_window_width))
    study = study.resample(time=resample_freq.pandas_freq)
    if date_event:
        return _reduce_with_date_event(
            resampled=study,
            reducer=resampled_op,
            window=rolling_window_width,
            source_delta=source_freq_delta,
        )
    return resampled_op(study, dim="time")


def _run_simple_reducer(
    climate_vars: list[ClimateVariable],
    resample_freq: Frequency,
    reducer_op: Callable[..., DataArray],
    date_event: bool,
    must_convert_rate: bool = False,
) -> DataArray:
    """
    Apply a simple reducer operation on climate variables.

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        List of climate variables to be processed.
    resample_freq : Frequency
        Frequency at which the data should be resampled.
    reducer_op : Callable[..., DataArray]
        Reducer operation to be applied on the data.
    date_event : bool
        Flag indicating whether the date when the event occurred should be added
        as a coordinate variable.
        Only works for `max` and `min` reducers.
        Defaults to False.
    must_convert_rate : bool, optional
        Flag indicating whether the data should be converted from rate to amount.
        Defaults to False.

    Returns
    -------
    DataArray
        Result of the reducer operation applied on the climate variables.
    """
    study, threshold = get_single_var(climate_vars)
    if threshold is not None:
        exceedance = _compute_exceedance(
            study=study,
            freq=resample_freq.pandas_freq,
            threshold=threshold,
            bootstrap=must_run_bootstrap(study, threshold),
        ).squeeze()
        filtered_study = study.where(exceedance)
    else:
        filtered_study = study
    if must_convert_rate and _is_rate(filtered_study):
        filtered_study = rate2amount(filtered_study)
    if date_event:
        return _reduce_with_date_event(
            resampled=filtered_study.resample(time=resample_freq.pandas_freq),
            reducer=reducer_op,
        )
    return reducer_op(
        filtered_study.resample(time=resample_freq.pandas_freq),
        dim="time",
    )


def get_single_var(
    climate_vars: list[ClimateVariable],
) -> tuple[DataArray, Threshold | None]:
    """
    Get the single variable and its threshold (if available).

    Parameters
    ----------
    climate_vars : list[ClimateVariable]
        A list of ClimateVariable objects.

    Returns
    -------
    tuple[DataArray, Threshold | None]
        A tuple containing the single variable's data array and its threshold
        (if available).
    """
    if climate_vars[0].threshold:
        return (
            climate_vars[0].studied_data,
            climate_vars[0].threshold,
        )
    return climate_vars[0].studied_data, None


def _compute_exceedances(
    climate_vars: list[ClimateVariable],
    resample_freq: str,
    logical_link: LogicalLink,
) -> DataArray:
    exceedances = [
        _compute_exceedance(
            study=climate_var.studied_data,
            threshold=climate_var.threshold,
            freq=resample_freq,
            bootstrap=must_run_bootstrap(
                climate_var.studied_data,
                climate_var.threshold,
            ),
        ).squeeze()
        for climate_var in climate_vars
    ]
    return logical_link(exceedances)


def _get_thresholded_var(
    climate_vars: list[ClimateVariable],
) -> tuple[DataArray, Threshold]:
    if climate_vars[0].threshold:
        return (
            climate_vars[0].studied_data,
            climate_vars[0].threshold,
        )
    msg = "No threshold found"
    raise InvalidIcclimArgumentError(msg)


def _reduce_with_date_event(
    resampled: DataArrayResample,
    reducer: Callable[[DataArrayResample], DataArray],
    source_delta: timedelta | None = None,
    window: int | None = None,
) -> DataArray:
    acc: list[DataArray] = []
    if reducer == DataArrayResample.max:
        group_reducer = DataArray.argmax
    elif reducer == DataArrayResample.min:
        group_reducer = DataArray.argmin
    else:
        msg = f"Can't compute `date_event` due to unknown reducer: '{reducer}'"
        raise NotImplementedError(msg)
    for label, sample in resampled:
        reduced_result = sample.isel(time=group_reducer(sample, dim="time"))
        if window is not None:
            result = _add_date_coords(
                original_sample=sample,
                result=sample.sum(dim="time"),
                start_time=reduced_result.time,
                end_time=reduced_result.time + window * source_delta,
                label=label,
            )
        else:
            result = _add_date_coords(
                original_sample=sample,
                result=sample.sum(dim="time"),
                event_date=reduced_result.time,
                label=label,
            )
        acc.append(result)
    return xr.concat(acc, "time")


def _count_occurrences_with_date(resampled: DataArrayResample) -> DataArray:
    acc: list[DataArray] = []
    for label, _sample in resampled:
        # TODO @bzah: probably not safe to compute on huge dataset,
        #              it should be fixed with
        #  https://github.com/pydata/xarray/issues/2511
        sample = _sample.compute()
        first = sample.isel(time=sample.argmax("time")).time
        reversed_time = sample.reindex(time=list(reversed(sample.time.to_numpy())))
        last = reversed_time.isel(time=reversed_time.argmax("time")).time
        dated_occurrences = _add_date_coords(
            original_sample=sample,
            result=sample.sum(dim="time"),
            start_time=first,
            end_time=last,
            label=label,
        )
        acc.append(dated_occurrences)
    return xr.concat(acc, "time")


def _consecutive_occurrences_with_dates(
    resampled: DataArrayResample,
    source_freq_delta: timedelta,
) -> DataArray:
    acc = []
    for label, _sample in resampled:
        sample = _sample.where(~_sample.isnull(), 0)
        time_index_of_max_rle = sample.argmax(dim="time")
        # TODO @bzah: `.compute` is needed until xarray merges this pr:
        # https://github.com/pydata/xarray/pull/5873
        time_index_of_max_rle = time_index_of_max_rle.compute()
        dated_longest_run = sample[{"time": time_index_of_max_rle}]
        start_time = sample.isel(
            time=time_index_of_max_rle.where(time_index_of_max_rle > 0, 0),
        ).time
        end_time = start_time + (dated_longest_run * source_freq_delta)
        dated_longest_run = _add_date_coords(
            original_sample=sample,
            result=dated_longest_run,
            start_time=start_time,
            end_time=end_time,
            label=label,
        )
        acc.append(dated_longest_run)
    return xr.concat(acc, "time")


def _add_date_coords(
    original_sample: DataArray,
    result: DataArray,
    label: str | np.datetime64,
    start_time: DataArray = None,
    end_time: DataArray = None,
    event_date: DataArray = None,
) -> DataArray:
    new_coords = {c: original_sample[c] for c in original_sample.coords if c != "time"}
    if event_date is None:
        new_coords["event_date_start"] = start_time
        new_coords["event_date_end"] = end_time
    else:
        new_coords["event_date"] = event_date
    new_coords["time"] = label
    return DataArray(data=result, coords=new_coords)


def _to_percent(da: DataArray, sampling_freq: Frequency) -> DataArray:
    if sampling_freq == FrequencyRegistry.MONTH:
        da = da / da.time.dt.daysinmonth * 100
    elif sampling_freq == FrequencyRegistry.YEAR:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[{"time": leap_years}] = 366
        coef[{"time": ~leap_years}] = 365
        da = da / coef
    elif sampling_freq == FrequencyRegistry.AMJJAS:
        da = da / 183
    elif sampling_freq == FrequencyRegistry.ONDJFM:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[{"time": leap_years}] = 183
        coef[{"time": ~leap_years}] = 182
        da = da / coef
    elif sampling_freq == FrequencyRegistry.DJF:
        coef = xr.full_like(da, 1)
        leap_years = _is_leap_year(da)
        coef[{"time": leap_years}] = 91
        coef[{"time": ~leap_years}] = 90
        da = da / coef
    elif sampling_freq in [FrequencyRegistry.MAM, FrequencyRegistry.JJA]:
        da = da / 92
    elif sampling_freq == FrequencyRegistry.SON:
        da = da / 91
    else:
        # TODO @bzah: improve this for custom resampling
        # https://github.com/cerfacs-globc/icclim/issues/289
        warn(
            "For now, '%' unit can only be used when `slice_mode` is one of: "
            "{MONTH, YEAR, AMJJAS, ONDJFM, DJF, MAM, JJA, SON}.",
            stacklevel=2,
        )
        return da
    da.attrs[UNITS_KEY] = PART_OF_A_WHOLE_UNIT
    return da


def _is_leap_year(da: DataArray) -> np.ndarray:
    time_index = da.indexes.get("time")
    if isinstance(time_index, xr.CFTimeIndex):
        return CfCalendarRegistry.lookup(time_index.calendar).is_leap(da.time.dt.year)
    return da.time.dt.is_leap_year


def _is_rate(query: Quantity | DataArray) -> bool:
    if isinstance(query, DataArray):
        query = str2pint(query.attrs[UNITS_KEY])
    return query.dimensionality.get("[time]", None) == -1


def check_freq(da, dim: str = "time", strict: bool = True):
    """
    Infer the sampling frequency of a DataArray along a given dimension.

    For daily climate data, always returns "D", even after seasonal slicing
    (which can break pandas' infer_freq). Falls back to pandas.infer_freq
    if not daily.
    """
    times = da[dim].values
    if len(times) < 2:
        return None

    # Compute deltas
    try:
        deltas = to_timedelta(np_diff(times))
    except Exception as e:
        if strict:
            raise ValueError(f"[icclim] Cannot compute time deltas: {e}")
        return None

    median_delta = Timedelta(np_median(deltas))

    # If ~daily, force "D"
    if np_abs(median_delta / Timedelta("1D") - 1) < 1e-6:
        return "D"

    # Otherwise, try pandas inference
    try:
        return infer_freq(
            date_range(start=times[0], periods=len(times), freq=median_delta)
        )
    except Exception as e:
        if strict:
            raise ValueError(f"[icclim] Unable to infer frequency: {e}")
        return None



"""Module to parse input data and make it usable for icclim."""

from __future__ import annotations

import re
import warnings
from collections.abc import Sequence
from datetime import datetime
from typing import TYPE_CHECKING

import numpy as np
import xarray as xr
import xclim
from xarray.core.dataarray import DataArray
from xarray.core.dataset import Dataset
from xclim.core.units import convert_units_to

from icclim._core.constants import UNITS_KEY, VALID_PERCENTILE_DIMENSION
from icclim._core.model.cf_calendar import CfCalendarRegistry
from icclim._core.model.standard_variable import (
    StandardVariable,
    StandardVariableRegistry,
)
from icclim._core.utils import read_date
from icclim.exception import InvalidIcclimArgumentError

if TYPE_CHECKING:
    from collections.abc import Hashable
    from datetime import datetime

    import pint
    from pint import Quantity

    from icclim._core.model.icclim_types import InFileBaseType, InFileLike
    from icclim._core.model.in_file_dictionary import InFileDictionary
    from icclim._core.model.standard_index import StandardIndex
    from icclim._core.model.threshold import Threshold


DEFAULT_INPUT_FREQUENCY = "days"
PR_AMOUNT_STANDARD_NAME = "thickness_of_rainfall_amount"


class PercentileDataArray(xr.DataArray):
    """Wrap xarray DataArray for percentiles values."""

    __slots__ = ()

    @classmethod
    def is_compatible(cls, source: xr.DataArray) -> bool:
        """Evaluate whether PecentileDataArray is conformant with expected fields.

        A PercentileDataArray must have climatology_bounds attributes and either a
        quantile or percentiles coordinate, the window is not mandatory.
        """
        return (
            isinstance(source, xr.DataArray)
            and source.attrs.get("climatology_bounds", None) is not None
            and ("quantile" in source.coords or "percentiles" in source.coords)
        )

    @classmethod
    def from_da(
        cls,
        source: xr.DataArray,
        climatology_bounds: list[str] | None = None,
    ) -> PercentileDataArray:
        """Create a PercentileDataArray from a xarray.DataArray.

        Parameters
        ----------
        source : xr.DataArray
            A DataArray with its content containing percentiles values.
            It must also have a coordinate variable percentiles or quantile.
        climatology_bounds : list[str]
            Optional. A List of size two which contains the period on which the
            percentiles were computed. See
            `xclim.core.calendar.build_climatology_bounds`
            to build this list from a DataArray.

        Returns
        -------
        PercentileDataArray
            The initial `source` DataArray but wrap by PercentileDataArray class.
            The data is unchanged and only climatology_bounds attributes is overridden
            if q new value is given in inputs.
        """
        if (
            climatology_bounds is None
            and source.attrs.get("climatology_bounds", None) is None
        ):
            msg = "PercentileDataArray needs a climatology_bounds."
            raise ValueError(msg)
        per = cls(source)
        # handle case where da was created with `quantile()` method
        if "quantile" in source.coords:
            per = per.rename({"quantile": "percentiles"})
            per.coords["percentiles"] = per.coords["percentiles"] * 100
        clim_bounds = source.attrs.get("climatology_bounds", climatology_bounds)
        per.attrs["climatology_bounds"] = clim_bounds
        if "percentiles" in per.coords:
            return per
        msg = (
            f"DataArray {source.name} could not be turned into"
            f" PercentileDataArray. The DataArray must have a"
            f" 'percentiles' coordinate variable."
        )
        raise ValueError(msg)


def guess_var_names(
    ds: Dataset,
    var_names: str | Sequence[str] | None,
    standard_index: StandardIndex | None,
) -> list[Hashable]:
    """
    Attempt to guess the variable names from the dataset and the standard index.

    Parameters
    ----------
    ds : Dataset
        The dataset to guess the variable names from.
    var_names : str | Sequence[str] | None
        The variable names to use. If None, the function will attempt to guess the
        variable names.
    standard_index : StandardIndex | None
        The standard index to use to guess the variable names.

    Returns
    -------
    list[Hashable]
        The list of guessed variable names.
    """
    if var_names is None:
        return _guess_dataset_var_names(ds=ds, standard_index=standard_index)
    if isinstance(var_names, str):
        return [var_names]
    if isinstance(var_names, (list, tuple)):
        return var_names
    msg = "`var_name` must be a string a list or None."
    raise NotImplementedError(msg)


def read_dataset(
    in_files: InFileBaseType,
    standard_var: StandardVariable | None = None,
    var_name: str | Sequence[str] | None = None,
) -> Dataset:
    """
    Read a dataset from input files.

    Parameters
    ----------
    in_files : InFileBaseType
        The input files to read the dataset from. It can be a single file path,
        a list of file paths, a glob pattern, a netCDF file, or a Zarr store.
    standard_var : StandardVariable | None, optional
        The standard variable to use for parsing the dataset, by default None.
    var_name : str | Sequence[str] | None, optional
        The variable name(s) to extract from the dataset, by default None.

    Returns
    -------
    Dataset
        The parsed dataset.

    Raises
    ------
    NotImplementedError
        If the format of `in_files` is not recognized.

    Notes
    -----
    This function supports reading datasets from various file formats, including
    netCDF and Zarr. It can handle single files, multiple files, and glob patterns.
    If `standard_var` is provided, it will use the specified standard variable for
    parsing the dataset. If `var_name` is provided, it will extract the specified
    variable(s) from the dataset.

    Examples
    --------
    >>> files = ["data1.nc", "data2.nc"]
    >>> ds = read_dataset(files, standard_var="temperature", var_name="temp")
    """
    if isinstance(in_files, Dataset):
        ds = in_files
    elif isinstance(in_files, DataArray):
        ds = _read_dataarray(in_files, standard_var=standard_var, var_name=var_name)
    elif is_glob_path(in_files) or (
        isinstance(in_files, (list, tuple)) and is_netcdf_path(in_files[0])
    ):
        # we assumes it's a list of netCDF files
        #  join="override" is used for cases some dimension are a tiny bit different
        #  in different files (was the case with eobs).
        # TODO @bzah: change parallel to True when issue is fixed on netcdf4 (py and C)
        # https://github.com/Unidata/netcdf4-python/issues/1192
        ds = xr.open_mfdataset(in_files, parallel=False, join="override")
    elif is_netcdf_path(in_files):
        ds = xr.open_dataset(in_files)
    elif is_zarr_path(in_files):
        ds = xr.open_zarr(in_files)
    elif isinstance(in_files, (list, tuple)):
        return xr.merge(
            [
                read_dataset(in_file, standard_var, var_name[i])
                for i, in_file in enumerate(in_files)
            ],
        )
    else:
        msg = f"`in_files` format {type(in_files)} was not recognized."
        raise NotImplementedError(msg)
    return update_to_standard_coords(ds)


def update_to_standard_coords(ds: Dataset) -> Dataset:
    """Mutate input ds to use more icclim friendly coordinate names."""
    # TODO @bzah: see if cf-xarray could replace this
    # https://github.com/cerfacs-globc/icclim/issues/289
    if ds.coords.get("t") is not None:
        ds = ds.rename({"t": "time"})
    return ds


def is_zarr_path(path: InFileBaseType) -> bool:
    """Check if the input path is a Zarr store."""
    return isinstance(path, str) and ".zarr" in path


def is_netcdf_path(path: InFileBaseType) -> bool:
    """Check if the input path is a netCDF file."""
    return isinstance(path, str) and ".nc" in path


def is_glob_path(path: InFileBaseType) -> bool:
    """Check if the input path is a glob pattern."""
    return isinstance(path, str) and "*" in path


def standardize_percentile_dim_name(per_da: DataArray) -> DataArray:
    """
    Standardizes the name of the percentile dimension in the input DataArray.

    Parameters
    ----------
    per_da : DataArray
        The input DataArray containing percentile data.

    Returns
    -------
    DataArray
        The input DataArray with the percentile dimension standardized.

    Raises
    ------
    InvalidIcclimArgumentError
        If the percentile data does not contain a recognizable percentiles dimension.

    Notes
    -----
    This function standardizes the name of the percentile dimension in the input
    DataArray to "percentiles".

    If the percentile dimension name contains the word "quantile", the values in the
    "percentiles" coordinate are multiplied by 100.
    """
    per_dim_name = None
    for d in VALID_PERCENTILE_DIMENSION:
        if d in per_da.dims:
            per_dim_name = d
        elif f"{d}s" in per_da.dims:
            # plural handling
            per_dim_name = f"{d}s"
    if per_dim_name is None:
        msg = (
            "Percentile data must contain a recognizable percentiles dimension such as"
            " 'percentiles', 'quantile', 'per' or 'centile'."
        )
        raise InvalidIcclimArgumentError(msg)
    per_da = per_da.rename({per_dim_name: "percentiles"})
    if "quantile" in per_dim_name:
        per_da.coords["percentiles"] = per_da.coords["percentiles"] * 100
    return per_da


def get_date_to_iso_format(in_date: str | datetime) -> str:
    """
    Get a date in ISO format from a string or a datetime object.

    Parameters
    ----------
    in_date: str | datetime
        A string representing a date or a datetime object.

    Returns
    -------
    str
        A string representing a date in ISO format.
    """
    if isinstance(in_date, str):
        if re.match(r"^\d{4}$", in_date):
            warnings.warn(
                f"{in_date} is transformed into {in_date}-01-01",
                stacklevel=2,
            )
            in_date += "-01-01"
        if re.match(r"^\d{4}-\d{2}$", in_date):
            warnings.warn(f"{in_date} is transformed into {in_date}-01", stacklevel=2)
            in_date += "-01"
        in_date = read_date(in_date)
    return in_date.strftime("%Y-%m-%d")


def read_clim_bounds(
    climatology_bounds: Sequence[str, str] | None,
    per_da: xr.DataArray,
) -> list[str]:
    """
    Read climatology bounds from input.

    The climatology bounds represent the start and end dates of the climatology period.

    Parameters
    ----------
    climatology_bounds : sequence of str or None
        The climatology bounds as a sequence of two strings representing the start and
        end dates.
        If None, the climatology bounds will be retrieved from the `climatology_bounds`
        attribute of `per_da`.
    per_da : xr.DataArray
        The input data array.

    Returns
    -------
    list of str
        A list of climatology bounds converted to ISO format.

    Raises
    ------
    InvalidIcclimArgumentError
        If the length of `climatology_bounds` is not equal to 2.

    Notes
    -----
    If `climatology_bounds` is None, the function will attempt to retrieve the
    climatology bounds from the `climatology_bounds` attribute of `per_da`.
    """
    bds = climatology_bounds or per_da.attrs.get("climatology_bounds", None)
    if len(bds) != 2:
        msg = "climatology_bounds must be an iterable of length 2."
        raise InvalidIcclimArgumentError(msg)
    return [get_date_to_iso_format(bd) for bd in bds]


def build_input_dict(
    in_files: InFileLike,
    var_names: Sequence[str] | None,
    threshold: Threshold | Sequence[Threshold] | None,
    standard_index: StandardIndex | None,
) -> dict[str, InFileDictionary]:
    """
    Build an input dictionary based on the provided input files and variables.

    The input dictionary is used to map which input files correspond to which variables.

    Parameters
    ----------
    in_files : InFileLike
        The input files. It can be a dictionary where the keys represent the variable
        names and the values represent the file paths, or a single file path.
    var_names : Sequence[str] | None
        The variable names. If `in_files` is a dictionary, this parameter must be None.
        Otherwise, it should be a sequence of variable names corresponding to the single
        file path.
    threshold : Threshold | Sequence[Threshold] | None
        The threshold values. It can be a single threshold value, a sequence of
        threshold values, or None.
    standard_index : StandardIndex | None
        The standard index. It can be a standard index value or None.

    Returns
    -------
    dict[str, InFileDictionary]
        The built input dictionary.

    Raises
    ------
    InvalidIcclimArgumentError
        If `var_names` is not None when `in_files` is a dictionary.

    Notes
    -----
    - If `in_files` is a dictionary, the dictionary keys are used as variable names.
    - If `in_files` is a dictionary and the dictionary values are also dictionaries,
      the input dictionary is returned as is.
    - If `in_files` is a dictionary and the dictionary values are file paths,
      the input dictionary is built using the file paths and variable names.
    - If `in_files` is a single file path and `var_names` is a single variable name,
      the input dictionary is built using the file path and variable name.
    """
    if isinstance(in_files, dict):
        if var_names is not None:
            msg = (
                "`var_name` must be None when `in_files` is a dictionary."
                " The dictionary keys are used in place of `var_name`."
            )
            raise InvalidIcclimArgumentError(msg)
        if isinstance(next(iter(in_files.values())), dict):
            # case of in_files={tasmax: {"study": "tasmax.nc"}}
            return in_files
        # case of in_files={tasmax: "tasmax.nc"}
        return _build_in_file_dict(
            in_files=list(in_files.values()),
            standard_index=standard_index,
            threshold=threshold,
            var_names=list(in_files.keys()),
        )
    # case of in_files="tasmax.nc" and var_names="tasmax"
    return _build_in_file_dict(in_files, var_names, threshold, standard_index)


def find_standard_vars(ds: Dataset) -> list[Hashable]:
    """
    Find standard variables in a dataset.

    Parameters
    ----------
    ds : Dataset
        The input dataset.

    Returns
    -------
    list[Hashable]
        A list of standard variables found in the dataset.
    """
    return [
        v
        for v in ds.data_vars
        if StandardVariableRegistry.lookup_no_error(str(v)) is not None
    ]


def guess_standard_variable(data: DataArray) -> StandardVariable | None:
    """
    Guesses the standard variable based on the metadata of `data`.

    Parameters
    ----------
    data : DataArray
        The input data.

    Returns
    -------
    StandardVariable or None
        The guessed standard variable, or None if no standard variable is found.

    Notes
    -----
    StandardVariableRegistry is used as a lookup table to find the standard variable
    using the dataarray name or standard name attribute.
    """
    std_var = StandardVariableRegistry.lookup_no_error(str(data.name))
    if std_var is None and data.attrs.get("standard_name", None) is not None:
        std_var = StandardVariableRegistry.lookup(
            data.attrs.get("standard_name"),
            no_error=True,
        )
    if std_var is None:
        return None
    return std_var


def is_precipitation_amount(source: xr.DataArray) -> bool:
    """
    Return True if the source is a precipitation amount.

    Parameters
    ----------
    source: xr.DataArray
        A DataArray object.

    Returns
    -------
    bool
        True if the source is a precipitation amount, False otherwise.

    Notes
    -----
    Using pint, the rate is a quantity with a dimensionality of [time]^-1.
    """
    standard_name = source.attrs.get("standard_name", None)
    source_unit = xclim.core.units.units2pint(source)
    return standard_name == PR_AMOUNT_STANDARD_NAME and _is_amount(source_unit)


def build_studied_data(
    original_da: DataArray,
    time_range: Sequence[datetime | str] | None,
    ignore_Feb29th: bool,  # noqa: N803
    default_units: str | None,
) -> DataArray:
    """
    Preprocesss the input data to select the period of interest.

    Parameters
    ----------
    original_da : DataArray
        The original data array.
    time_range : Sequence[datetime | str] | None
        The time range to select from the data array. If None, the entire time range is
        used.
    ignore_Feb29th : bool
        Whether to ignore February 29th when processing the data.
    default_units : str | None
        The default units to use for the data array if it is uniteless.
        If None and the data array is uniteless, "units" attribute remains unset.

    Returns
    -------
    DataArray
        The processed data array.

    Raises
    ------
    InvalidIcclimArgumentError
        If the given `time_range` is out of the dataset time period.

    """
    if time_range is not None:
        _check_time_range_pre_validity("time_range", time_range)
        time_range = [get_date_to_iso_format(x) for x in time_range]
        da = original_da.sel(time=slice(time_range[0], time_range[1]))
        _check_time_range_post_validity(da, original_da, "time_range", time_range)
        if len(da.time) == 0:
            msg = (
                f"The given `time_range` {time_range} is out of the dataset time"
                f" period: {original_da.time.min().dt.floor('D').values}"
                f" - {original_da.time.max().dt.floor('D').values}."
            )
            raise InvalidIcclimArgumentError(msg)
    else:
        da = original_da
    if ignore_Feb29th:
        da = xclim.core.calendar.convert_calendar(da, CfCalendarRegistry.NO_LEAP.name)
    if da.attrs.get(UNITS_KEY, None) is None and default_units is not None:
        da.attrs[UNITS_KEY] = default_units
    if is_precipitation_amount(da):
        da = xclim.core.units.amount2rate(da)
    return da.chunk("auto")


def get_name_of_first_var(ds: Dataset) -> str:
    """
    Get the name of the first variable in the given Dataset.

    Parameters
    ----------
    ds : Dataset
        The input Dataset.

    Returns
    -------
    str
        The name of the first variable in the Dataset.

    Raises
    ------
    IndexError
        If the Dataset is empty.
    """
    return str(ds.data_vars[next(iter(ds.data_vars.keys()))].name)


def is_dataset_path(query: InFileBaseType) -> bool:
    """
    Check if the given query is a valid dataset path.

    Parameters
    ----------
    query : InFileBaseType
        The query to check. It can be a single path or a list/tuple of paths.

    Returns
    -------
    bool
        True if the query is a valid dataset path, False otherwise.

    Notes
    -----
    A valid dataset path can be either a NetCDF path, a Zarr path, a glob path, or a
    list/tuple of valid paths.

    """
    if isinstance(query, (tuple, list)):
        return all(is_netcdf_path(q) for q in query)
    return is_zarr_path(query) or is_glob_path(query) or is_netcdf_path(query)


def build_reference_da(
    original_da: DataArray,
    base_period_time_range: Sequence[datetime | str] | None,
    only_leap_years: bool,
    percentile_min_value: Quantity | None,
) -> DataArray:
    """
    Build a reference DataArray to be used for percentile doy computation.

    Parameters
    ----------
    original_da: DataArray
        The DataArray used as a base.
    base_period_time_range: Sequence[datetime | str] | None
        The period to slice in the base DataArray.
    only_leap_years: bool
        Flag to only use leap years (years with 366 days).
    percentile_min_value: Quantity | None
        Optional, if set will replace every value from the base DataArray that are below
        the `percentile_min_value` with np.nan.
    """
    reference = original_da
    if base_period_time_range:
        _check_time_range_pre_validity("base_period_time_range", base_period_time_range)
        base_period_time_range = [
            get_date_to_iso_format(x) for x in base_period_time_range
        ]
        reference = original_da.sel(
            time=slice(base_period_time_range[0], base_period_time_range[1]),
        )
        _check_time_range_post_validity(
            reference,
            original_da,
            "base_period_time_range",
            base_period_time_range,
        )
    if only_leap_years:
        reference = _reduce_only_leap_years(original_da)
    if percentile_min_value is not None:
        percentile_min_value = convert_units_to(
            str(percentile_min_value),
            reference,
            context="hydro",
        )
        reference = reference.where(reference >= percentile_min_value, np.nan)
    return reference


def get_dataarray_from_dataset(
    var_name: str | None,
    value: xr.Dataset | str,
    standard_var: StandardVariable | None = None,
) -> xr.DataArray:
    """
    Extract a DataArray from a Dataset based on the provided variable name.

    Parameters
    ----------
    var_name : str or None
        The name of the variable to extract from the Dataset. If None, the function
        will try to guess the variable based on the Dataset's contents.
    value : xr.Dataset or str
        The Dataset object or the path to the Dataset file.
    standard_var : StandardVariable
        The standard variable used to find a matching variable in the Dataset.

    Returns
    -------
    xr.DataArray
        The extracted DataArray.

    Raises
    ------
    InvalidIcclimArgumentError
        If the variable name cannot be guessed and `var_name` is None.

    Notes
    -----
    This function can be used to extract a specific variable from a Dataset object
    or a Dataset file. If `var_name` is None, the function will try to guess the
    variable based on the Dataset's contents.
    """
    ds = (
        value
        if isinstance(value, xr.Dataset)
        else read_dataset(value, standard_var=standard_var)
    )
    if var_name is not None:
        return ds[var_name]
    if len(ds.data_vars) == 1:
        return ds[get_name_of_first_var(ds)]
    matching_name = _find_matching_standard_var(ds, standard_var)
    if matching_name is not None:
        return ds[matching_name]
    names = find_standard_vars(ds)
    if len(names) == 1:
        return ds[var_name]
    msg = (
        f"Could not guess the variable to use for the normal in {ds}."
        f" Use `normal_var_name` to specify which variable should be"
        f" used."
    )
    raise InvalidIcclimArgumentError(msg)


def _build_in_file_dict(
    in_files: InFileBaseType,
    var_names: Sequence[str],
    threshold: Threshold | Sequence[Threshold] | None,
    standard_index: StandardIndex | None,
) -> InFileDictionary:
    standard_var = (
        standard_index.input_variables[0] if standard_index is not None else None
    )
    input_dataset = read_dataset(
        in_files=in_files,
        standard_var=standard_var,
        var_name=var_names,
    )
    var_names = guess_var_names(
        ds=input_dataset,
        standard_index=standard_index,
        var_names=var_names,
    )
    if threshold is not None:
        if len(var_names) == 1:
            return {
                var_names[0]: {
                    "study": input_dataset[var_names[0]],
                    "thresholds": threshold,
                },
            }
        if not isinstance(threshold, Sequence):
            threshold = [threshold]
        if len(threshold) != len(var_names):
            # Allow 1 var with multiple thresholds or 1 threshold per var
            # but no other case
            msg = (
                "There must be as many thresholds as there are variables. There was"
                f" {len(threshold)} thresholds and {len(var_names)} variables."
            )
            raise InvalidIcclimArgumentError(msg)
        return {
            var_name: {"study": input_dataset[var_name], "thresholds": threshold[i]}
            for i, var_name in enumerate(var_names)
        }
    return {var_name: {"study": input_dataset[var_name]} for var_name in var_names}


def _read_dataarray(
    data: DataArray,
    standard_var: StandardVariable | None = None,
    var_name: str | Sequence[str] | None = None,
) -> Dataset:
    if isinstance(var_name, (tuple, list)):
        if len(var_name) > 1:
            msg = (
                "When `in_file` is a DataArray, there"
                f" can only be one value in `var_name` but var_name was: {var_name} "
            )
            raise InvalidIcclimArgumentError(msg)
        var_name = var_name[0]
        data_name = var_name or standard_var.short_name or None
    else:
        data_name = var_name or data.name or "unnamed_var"
    return data.to_dataset(name=data_name, promote_attrs=True)


def _guess_dataset_var_names(
    standard_index: StandardIndex,
    ds: Dataset,
) -> list[Hashable]:
    """Try to guess the variable names.

    The expected kind of variable of the index is used to guess the variable names.
    """
    if standard_index is not None:
        main_aliases = ", ".join(
            (v.short_name for v in standard_index.input_variables),
        )
        error_msg = (
            f"Index {standard_index.short_name} needs the following variable(s)"
            f" [{main_aliases}], but the input variables were {list(ds.data_vars)}."
            f" Use `var_name` parameter to explicitly set variable names."
        )
        if len(ds.data_vars) == 1:
            if len(standard_index.input_variables) != 1:
                raise InvalidIcclimArgumentError(error_msg)
            return [get_name_of_first_var(ds)]
        climate_var_names = []
        for expected_standard_var in standard_index.input_variables:
            for alias in expected_standard_var.aliases:
                # check if dataset contains this alias
                if _is_alias_valid(ds, alias):
                    climate_var_names.append(_get_actual_name(ds, alias))
                    break
        if len(climate_var_names) < len(standard_index.input_variables):
            raise InvalidIcclimArgumentError(error_msg)
        return climate_var_names
    if len(ds.data_vars) == 1:
        return [get_name_of_first_var(ds)]
    return find_standard_vars(ds)


def _check_time_range_pre_validity(key: str, tr: Sequence[datetime | str]) -> None:
    if len(tr) != 2:
        msg = (
            f"The given `{key}` {tr}"
            f" has {len(tr)} elements."
            f" It must have exactly 2 dates."
        )
        raise InvalidIcclimArgumentError(msg)


def _check_time_range_post_validity(
    da: DataArray, original_da: DataArray, key: str, tr: list
) -> None:
    if len(da.time) == 0:
        msg = (
            f"The given `{key}` {tr} is out of the sample time bounds:"
            f" {original_da.time.min().dt.floor('D').values}"
            f" - {original_da.time.max().dt.floor('D').values}."
        )
        raise InvalidIcclimArgumentError(msg)


def _is_alias_valid(ds: Dataset, alias: str) -> bool:
    return any(str(ds_var).upper() == alias.upper() for ds_var in ds.data_vars)


def _get_actual_name(ds: Dataset, alias: str) -> str:
    for ds_var in ds.data_vars:
        if str(ds_var).upper() == alias.upper():
            return str(ds_var)
    msg = f"Could not find {alias} in dataset."
    raise KeyError(msg)


def _reduce_only_leap_years(da: DataArray) -> DataArray:
    reduced_list = []
    for _, val in da.groupby(da.time.dt.year):
        if val.time.dt.dayofyear.max() == 366:
            reduced_list.append(val)
    if not reduced_list:
        msg = "No leap year in current dataset. Do not use `only_leap_years` parameter."
        raise InvalidIcclimArgumentError(msg)
    return xr.concat(reduced_list, "time")


def _is_rate(u: pint.Unit) -> bool:
    return u.dimensionality.get("[time]") == -1


def _is_amount(u: pint.Unit) -> bool:
    return not _is_rate(u)


def _find_matching_standard_var(ds: xr.Dataset, std: StandardVariable) -> str | None:
    for v in ds.data_vars:
        standardized_var = StandardVariableRegistry.lookup_no_error(str(v))
        if std == standardized_var:
            return str(v)
    return None



"""Contain icclim-specific exceptions."""

from __future__ import annotations


class InvalidIcclimArgumentError(ValueError):
    """
    Exception raised for erroneous input arguments.

    Attributes
    ----------
    msg : str
        Error description.
    source_err : Exception or None, optional
        The source of the error, if any.

    Methods
    -------
    __str__()
        Returns a string representation of the error message.
    """

    def __init__(self, msg: str, source_err: Exception | None = None) -> None:
        self.msg = msg
        self.source = source_err

    def __str__(self) -> str:
        """Return a string representation of the error message."""
        return repr(self.msg)



"""Proxy icclim indicators that redirect to xclim indicators."""

from __future__ import annotations

from typing import TYPE_CHECKING

import xclim
from pandas import to_datetime

from icclim._core.generic.functions import get_single_var
from icclim._core.generic.indicator import Indicator
from icclim.exception import InvalidIcclimArgumentError
from icclim.frequency import FrequencyRegistry

if TYPE_CHECKING:
    import xarray

    from icclim._core.model.index_config import IndexConfig


class GrowingSeasonLength(Indicator):
    """
    Proxy for xclim.growing_season_length.

    icclim indicator that redirect to xclim `growing_season_length` indicator.
    """

    name = "growing_season_length"
    standard_name = xclim.atmos.growing_season_length.standard_name
    long_name = "ECAD Growing Season Length (Tmean > 5 degree_Celsius)"
    cell_methods = ""

    def __call__(self, config: IndexConfig) -> xarray.DataArray:
        """Compute the growing season length."""
        study, _ = get_single_var(config.climate_variables)
        return xclim.atmos.growing_season_length(
            tas=study,
            thresh="5 degree_Celsius",
            window=6,
            mid_date="07-01",
            freq=config.frequency.pandas_freq,
        )

    def preprocess(self, *args, **kwargs) -> list[xarray.DataArray]:
        """Not implemented as xclim indicator already handle pre/post processing."""
        raise NotImplementedError

    def postprocess(self, *args, **kwargs) -> xarray.DataArray:
        """Not implemented as xclim indicator already handle pre/post processing."""
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        """Check if the other object is a GrowingSeasonLength instance."""
        return isinstance(other, GrowingSeasonLength)


class StandardizedPrecipitationIndex3(Indicator):
    """
    Proxy for xclim.atmos.standardized_precipitation_index.

    icclim indicator that redirect to xclim `standardized_precipitation_index`
    indicator, with 3 MS frequency preconfigured.
    """

    name = "standardized_precipitation_index_3"
    standard_name = xclim.atmos.standardized_precipitation_index.standard_name
    long_name = "3-Month Standardized Precipitation Index (SPI3)"
    cell_methods = ""

    def __call__(self, config: IndexConfig) -> xarray.DataArray:
        """Compute the 3-Month Standardized Precipitation Index."""
        if config.frequency is not FrequencyRegistry.YEAR:  # year is default freq
            msg = "`slice_mode` cannot be configured when computing SPI3"
            raise InvalidIcclimArgumentError(msg)
        study, _ = get_single_var(config.climate_variables)
        # Parse with correct format
        study_cv = config.climate_variables[0]  # the ClimateVariable
        if study_cv.reference_period is None:
            raise ValueError("reference_period is missing for SPI index")
        start, end = to_datetime(study_cv.reference_period, format="%m-%d-%Y")
        # Convert back to YYYY-MM-DD
        cal_start, cal_end = start.strftime("%Y-%m-%d"), end.strftime("%Y-%m-%d")
        return xclim.atmos.standardized_precipitation_index(
            pr=study,
            freq="MS",
            window=3,
            cal_start=cal_start,
            cal_end=cal_end,
            dist="gamma",
            method="APP",
            fitkwargs={"floc": 0},
        )

    def preprocess(self, *args, **kwargs) -> list[xarray.DataArray]:
        """Not implemented as xclim indicator already handle pre/post processing."""
        raise NotImplementedError

    def postprocess(self, *args, **kwargs) -> xarray.DataArray:
        """Not implemented as xclim indicator already handle pre/post processing."""
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        """Check if the other object is a StandardizedPrecipitationIndex3 instance."""
        return isinstance(other, StandardizedPrecipitationIndex3)


class StandardizedPrecipitationIndex6(Indicator):
    """
    Proxy for xclim.atmos.standardized_precipitation_index.

    icclim indicator that redirect to xclim `standardized_precipitation_index`
    indicator, with 6 MS configured.
    """

    name = "standardized_precipitation_index_6"
    standard_name = xclim.atmos.standardized_precipitation_index.standard_name
    long_name = "6-Month Standardized Precipitation Index (SPI6)"
    cell_methods = ""

    def __call__(self, config: IndexConfig) -> xarray.DataArray:
        """Compute the 6-Month Standardized Precipitation Index."""
        if config.frequency is not FrequencyRegistry.YEAR:  # year is default freq
            msg = "`slice_mode` cannot be configured when computing SPI6"
            raise InvalidIcclimArgumentError(msg)
        study, _ = get_single_var(config.climate_variables)
        # Parse with correct format
        study_cv = config.climate_variables[0]  # the ClimateVariable
        if study_cv.reference_period is None:
            raise ValueError("reference_period is missing for SPI index")
        start, end = to_datetime(study_cv.reference_period, format="%m-%d-%Y")
        # Convert back to YYYY-MM-DD
        cal_start, cal_end = start.strftime("%Y-%m-%d"), end.strftime("%Y-%m-%d")
        return xclim.atmos.standardized_precipitation_index(
            pr=study,
            freq="MS",
            window=6,
            cal_start=cal_start,
            cal_end=cal_end,
            dist="gamma",
            method="APP",
            fitkwargs={"floc": 0},
        )

    def preprocess(self, *args, **kwargs) -> list[xarray.DataArray]:
        """Not implemented as xclim indicator already handle pre/post processing."""
        raise NotImplementedError

    def postprocess(self, *args, **kwargs) -> xarray.DataArray:
        """Not implemented as xclim indicator already handle pre/post processing."""
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        """Check if the other object is a StandardizedPrecipitationIndex6 instance."""
        return isinstance(other, StandardizedPrecipitationIndex6)



# -*- Coding: latin-1 -*-
#  Copyright CERFACS (http://cerfacs.fr/)
#  Apache License, Version 2.0 (http://www.apache.org/licenses/LICENSE-2.0)
"""
Main entry point of icclim.

This module expose icclim principal function, notably `index` which is use by the
generated API.
A convenience function `indices` is also exposed to compute multiple indices at once.
"""

from __future__ import annotations

import datetime as dt
import operator
import time
from collections.abc import Callable, Sequence
from functools import reduce
from typing import TYPE_CHECKING
from warnings import warn

import numpy as np
import xarray as xr
import xclim

from icclim._core.climate_variable import (
    ClimateVariable,
    build_climate_vars,
)
from icclim._core.constants import (
    RESAMPLE_METHOD,
    UNITS_KEY,
)
from icclim._core.generic.indicator import GenericIndicator
from icclim._core.input_parsing import build_input_dict
from icclim._core.legacy.user_index import parse
from icclim._core.model.index_config import IndexConfig
from icclim._core.model.index_group import IndexGroup, IndexGroupRegistry
from icclim._core.model.logical_link import LogicalLinkRegistry
from icclim._core.model.netcdf_version import NetcdfVersion, NetcdfVersionRegistry
from icclim._core.model.quantile_interpolation import (
    QuantileInterpolation,
    QuantileInterpolationRegistry,
)
from icclim._core.model.standard_index import StandardIndex
from icclim._core.model.threshold import Threshold
from icclim._core.utils import read_date
from icclim.dcsc.registry import DcscIndexRegistry
from icclim.ecad.binding import (
    StandardizedPrecipitationIndex3,
    StandardizedPrecipitationIndex6,
)
from icclim.ecad.registry import EcadIndexRegistry
from icclim.exception import InvalidIcclimArgumentError
from icclim.frequency import Frequency, FrequencyRegistry
from icclim.generic.registry import GenericIndicatorRegistry
from icclim.logger import IcclimLogger, Verbosity, VerbosityRegistry
from icclim.threshold.factory import build_threshold

if TYPE_CHECKING:
    from xarray.core.dataarray import DataArray
    from xarray.core.dataset import Dataset

    from icclim._core.legacy.user_index.model import UserIndexDict
    from icclim._core.model.icclim_types import (
        FrequencyLike,
        InFileLike,
        SamplingMethodLike,
    )
    from icclim._core.model.in_file_dictionary import InFileDictionary
    from icclim._core.model.indicator import Indicator

log: IcclimLogger = IcclimLogger.get_instance(VerbosityRegistry.LOW)

HISTORY_CF_KEY = "history"
SOURCE_CF_KEY = "source"
ICCLIM_REFERENCE = "icclim"


def indices(
    index_group: Sequence[str] | str | IndexGroup | StandardIndex,
    *,
    ignore_error: bool = False,
    **kwargs,
) -> Dataset:
    """
    Compute multiple indices at the same time.

    The input dataset(s) must include all the necessary variables.
    It can only be used with keyword arguments (kwargs).

    Parameters
    ----------
    index_group : "all" | str | IndexGroup | list[str]
        Either the name of an IndexGroup or an instance of IndexGroup or a list
        of index short names or the name(s) of standard variable(s) (such as 'tasmax').
        The value "all" can also be used to compute every indices.
        Note that the input given by ``in_files`` must include all the necessary
        variables to compute the indices of this group.
    ignore_error: bool
        When True, ignore indices that fails to compute. This is option is particularly
        useful when used with `index_group='all'` to compute everything that can be
        computed given the input.
    kwargs : Dict
        ``icclim.index`` keyword arguments.

    Returns
    -------
    xr.Dataset
        A Dataset with one data variable per index.

    Notes
    -----
    If ``output_file`` is part of kwargs, the result is written in a single netCDF
    file, which will contain all the index results of this group.
    """
    indices = _get_ecad_indices_of_group(index_group)
    out = None
    if "out_file" in kwargs:
        out = kwargs["out_file"]
        del kwargs["out_file"]
    acc = []
    for i in indices:
        log.info("Computing index %s", i.short_name)
        kwargs["index_name"] = i.short_name
        if ignore_error:
            try:
                res = index(**kwargs)
                if "percentiles" in res.coords:
                    res = res.rename({"percentiles": i.short_name + "_percentiles"})
                if "thresholds" in res.coords:
                    res = res.rename({"thresholds": i.short_name + "_thresholds"})
                acc.append(res)
            except Exception:  # noqa: BLE001 (catch everything)
                warn(f"Could not compute {i.short_name}.", stacklevel=2)
        else:
            res = index(**kwargs)
            if "percentiles" in res.coords:
                res = res.rename({"percentiles": i.short_name + "_percentiles"})
            if "thresholds" in res.coords:
                res = res.rename({"thresholds": i.short_name + "_thresholds"})
            acc.append(res)
    ds: Dataset = xr.merge(acc)
    if out is not None:
        _write_output_file(
            result_ds=ds,
            input_time_encoding=ds.time.encoding,
            netcdf_version=kwargs.get("netcdf_version", NetcdfVersionRegistry.NETCDF4),
            file_path=out,
        )
    return ds


def _get_ecad_indices_of_group(
    query: Sequence[str] | str | IndexGroup | StandardIndex,
) -> list[StandardIndex]:
    if query == IndexGroupRegistry.WILD_CARD_GROUP or (
        isinstance(query, str)
        and query.lower() == IndexGroupRegistry.WILD_CARD_GROUP.name
    ):
        # case of group="all"
        return EcadIndexRegistry.values()
    if not isinstance(query, (list, tuple)):
        query = [query]
    # -- Look for standard indices (e.g. index_group='tx90p')
    indices = [EcadIndexRegistry.lookup_no_error(i) for i in query]
    indices = list(filter(lambda x: x is not None, indices))
    if len(indices) == len(query):
        return indices
    # -- Look for variables in standard indices (e.g. index_group='tasmax')
    indices = []
    for ecad_index in EcadIndexRegistry.values():
        has_var = True
        for var in ecad_index.input_variables:
            is_query_in_aliases = (
                standard_var in var.aliases for standard_var in query
            )
            has_var &= any(is_query_in_aliases)
        if has_var:
            indices.append(ecad_index)
    if len(indices) >= len(query):
        return indices
    # -- Look for index group (e.g. index_group='HEAT')
    groups = [IndexGroupRegistry.lookup_no_error(i) for i in query]
    groups = list(filter(lambda x: x is not None, groups))
    indices = (x.get_indices() for x in groups)
    indices = reduce(operator.add, indices, [])  # flatten list[list]
    if len(indices) >= len(query):
        return indices
    msg = f"The index group {query} was not recognized."
    raise InvalidIcclimArgumentError(msg)


def indice(*args, **kwargs) -> Dataset:
    """
    Proxy for `icclim.index` function.

    Deprecated: to be deleted in a future release.
    """
    log.deprecation_warning(old="icclim.indice", new="icclim.index")
    return index(*args, **kwargs)


def index(
    in_files: InFileLike,
    index_name: str
    | GenericIndicator
    | StandardIndex
    | None = None,  # optional when computing user_indices
    var_name: str | Sequence[str] | None = None,
    slice_mode: FrequencyLike | Frequency = "year",
    time_range: Sequence[dt.datetime | str] | None = None,
    out_file: str | None = None,
    threshold: str | Threshold | Sequence[str | Threshold] | None = None,
    callback: Callable[[int], None] = log.callback,
    callback_percentage_start_value: int = 0,
    callback_percentage_total: int = 100,
    base_period_time_range: Sequence[dt.datetime] | Sequence[str] | None = None,
    doy_window_width: int = 5,
    only_leap_years: bool = False,
    ignore_Feb29th: bool = False,  # noqa: N803
    interpolation: str | QuantileInterpolation = "median_unbiased",
    out_unit: str | None = None,
    netcdf_version: str | NetcdfVersion = "NETCDF4",
    user_index: UserIndexDict | None = None,
    save_thresholds: bool = False,
    logs_verbosity: Verbosity | str = "LOW",
    date_event: bool = False,
    min_spell_length: int | None = 6,
    rolling_window_width: int | None = 5,
    sampling_method: SamplingMethodLike = RESAMPLE_METHOD,
    *,
    # deprecated params are kwargs only
    window_width: int | None = None,
    save_percentile: bool | None = None,
    indice_name: str | None = None,
    user_indice: UserIndexDict | None = None,
    transfer_limit_Mbytes: float | None = None,  # noqa: N803
) -> Dataset:
    """
    Compute climate index.

    This is the main entry point for icclim.

    Parameters
    ----------
    in_files: str | list[str] | Dataset | DataArray | InputDictionary
        Absolute path(s) to NetCDF dataset(s), including OPeNDAP URLs,
        or path to zarr store, or xarray.Dataset or xarray.DataArray.
    index_name: str | StandardIndex
        Climate index name.
        For ECA&D index, case insensitive name used to lookup the index.
        For user index, it's the name of the output variable.
    var_name: str | list[str] | None
        ``optional`` Target variable name to process corresponding to ``in_files``.
        If None (default) on ECA&D index, the variable is guessed based on the
        climate index wanted.
        Mandatory for a user index.
    slice_mode: FrequencyLike | Frequency
        Type of temporal aggregation:
        The possibles values are ``{"year", "month", "DJF", "MAM", "JJA", "SON",
        "ONDJFM" or "AMJJAS", ("season", [1,2,3]), ("month", [1,2,3,])}``
        (where season and month lists can be customized) or any valid pandas
        frequency.
        A season can also be defined between two exact dates:
        ``("season", ("19 july", "14 august"))``.
        Default is "year".
        See :ref:`slice_mode` for details.
    time_range: list[datetime.datetime ] | list[str]  | tuple[str, str] | None
        ``optional`` Temporal range: upper and lower bounds for temporal subsetting.
        If ``None``, whole period of input files will be processed.
        The dates can either be given as instance of datetime.datetime or as string
        values. For strings, many format are accepted.
        Default is ``None``.
    out_file: str | None
        Output NetCDF file name (default: "icclim_out.nc" in the current directory).
        Default is "icclim_out.nc".
        If the input ``in_files`` is a ``Dataset``, ``out_file`` field is ignored.
        Use the function returned value instead to retrieve the computed value.
        If ``out_file`` already exists, icclim will overwrite it!
    threshold: float | list[float] | None
        ``optional`` User defined threshold for certain indices.
        Default depend on the index, see their individual definition.
        When a list of threshold is provided, the index will be computed for each
        thresholds.
    transfer_limit_Mbytes: float
        Deprecated, does not have any effect.
    callback: Callable[[int], None]
        ``optional`` Progress bar printing. If ``None``, progress bar will not be
        printed.
    callback_percentage_start_value: int
        ``optional`` Initial value of percentage of the progress bar (default: 0).
    callback_percentage_total: int
        ``optional`` Total percentage value (default: 100).
    base_period_time_range: list[datetime.datetime ] | list[str] | tuple[str, str] | None
        ``optional`` Temporal range of the reference period.
        The dates can either be given as instance of datetime.datetime or as string
        values.
        It is used either:
        #. to compute percentiles if threshold is filled.
        When missing, the studied period is used to compute percentiles.
        The study period is either the dataset filtered by `time_range` or the whole
        dataset if `time_range` is missing.
        For day of year percentiles (doy_per), on extreme percentiles the
        overlapping period between `base_period_time_range` and the study period is
        bootstrapped.
        #. to compute a reference period for indices such as difference_of_mean
        (a.k.a anomaly) if a single variable is given in input.
    doy_window_width: int
        ``optional`` Window width used to aggreagte day of year values when computing
        day of year percentiles (doy_per)
        Default: 5 (5 days).
    min_spell_length: int
        ``optional`` Minimum spell duration to be taken into account when computing
        the sum_of_spell_lengths.
    rolling_window_width: int
        ``optional`` Window width of the rolling window for indicators such as
        `{max_of_rolling_sum, max_of_rolling_average, min_of_rolling_sum, min_of_rolling_average}`
    only_leap_years: bool
        ``optional`` Option for February 29th (default: False).
    ignore_Feb29th: bool
        ``optional`` Ignoring or not February 29th (default: False).
    interpolation: str | QuantileInterpolation | None
        ``optional`` Interpolation method to compute percentile values:
        ``{"linear", "median_unbiased"}``
        Default is "median_unbiased", a.k.a type 8 or method 8.
        Ignored for non percentile based indices.
    out_unit: str | None
        ``optional`` Output unit for certain indices: "days" or "%"
        (default: "days").
    netcdf_version: str | NetcdfVersion
        ``optional`` NetCDF version to create (default: "NETCDF3_CLASSIC").
    user_index: UserIndexDict
        ``optional`` A dictionary with parameters for user defined index.
        See :ref:`Custom indices`.
        Ignored for ECA&D indices.
    save_thresholds: bool
        ``optional`` True if the thresholds should be saved within the resulting
        netcdf file (default: False).
    date_event: bool
        When True the date of the event (such as when a maximum is reached) will be
        stored in coordinates variables.
        **warning** This option may significantly slow down computation.
    logs_verbosity: str | Verbosity
        ``optional`` Configure how verbose icclim is.
        Possible values: ``{"LOW", "HIGH", "SILENT"}`` (default: "LOW")
    sampling_method: str
        Choose whether the output sampling configured in `slice_mode` is a
        `groupby` operation or a `resample` operation (as per xarray definitions).
        Possible values:
        ``{"groupby", "resample", "groupby_ref_and_resample_study"}``
        (default: "resample")
        `groupby_ref_and_resample_study` may only be used when computing the
        `difference_of_means` (a.k.a the anomaly).
    indice_name: str | None
        DEPRECATED, use index_name instead.
    user_indice: dict | None
        DEPRECATED, use user_index instead.
    window_width: int
        DEPRECATED, use doy_window_width, min_spell_length or rolling_window_width
        instead.
    save_percentile: bool
        DEPRECATED, use save_thresholds instead.

    """  # noqa: E501
    _setup(callback, callback_percentage_start_value, logs_verbosity)
    (
        index_name,
        user_index,
        save_thresholds,
        doy_window_width,
    ) = _handle_deprecated_params(
        index_name,
        user_index,
        save_thresholds,
        indice_name,
        transfer_limit_Mbytes,
        user_indice,
        save_percentile,
        window_width,
        doy_window_width,
    )
    del indice_name, transfer_limit_Mbytes, user_indice, save_percentile, window_width
    config = _build_config(
        in_files=in_files,
        index_name=index_name,
        var_name=var_name,
        slice_mode=slice_mode,
        time_range=time_range,
        threshold=threshold,
        callback=callback,
        base_period_time_range=base_period_time_range,
        doy_window_width=doy_window_width,
        only_leap_years=only_leap_years,
        ignore_Feb29th=ignore_Feb29th,
        interpolation=interpolation,
        out_unit=out_unit,
        netcdf_version=netcdf_version,
        user_index=user_index,
        save_thresholds=save_thresholds,
        date_event=date_event,
        min_spell_length=min_spell_length,
        rolling_window_width=rolling_window_width,
        sampling_method=sampling_method,
    )
    result_ds = _compute_climate_index(
        climate_index=config.indicator,
        config=config,
        initial_history=config.climate_variables[0].global_metadata["history"],
        initial_source=config.climate_variables[0].global_metadata["source"],
        rename=config.rename,
        reference=config.reference,
    )
    if out_file is not None:
        _write_output_file(
            result_ds,
            config.climate_variables[0].global_metadata["time_encoding"],
            config.netcdf_version,
            out_file,
        )
    callback(callback_percentage_total)
    log.ending_message(time.process_time())
    return result_ds


def _build_config(
    in_files: InFileLike,
    index_name: str | GenericIndicator | StandardIndex | None,
    var_name: str | Sequence[str] | None,
    slice_mode: FrequencyLike | Frequency,
    time_range: Sequence[dt.datetime | str] | None,
    threshold: str | Threshold | Sequence[str | Threshold] | None,
    callback: Callable[[int], None],
    base_period_time_range: Sequence[dt.datetime] | Sequence[str] | None,
    doy_window_width: int,
    only_leap_years: bool,
    ignore_Feb29th: bool,  # noqa: N803
    interpolation: str | QuantileInterpolation,
    out_unit: str | None,
    netcdf_version: str | NetcdfVersion,
    user_index: UserIndexDict | None,
    save_thresholds: bool,
    date_event: bool,
    min_spell_length: int | None,
    rolling_window_width: int | None,
    sampling_method: SamplingMethodLike,
) -> IndexConfig:
    if user_index is not None and (index_name is None or isinstance(index_name, str)):
        return _build_user_index_config(
            user_index,
            in_files=in_files,
            index_name=index_name,
            var_name=var_name,
            slice_mode=slice_mode,
            time_range=time_range,
            callback=callback,
            base_period_time_range=base_period_time_range,
            doy_window_width=doy_window_width,
            only_leap_years=only_leap_years,
            ignore_Feb29th=ignore_Feb29th,
            interpolation=interpolation,
            out_unit=out_unit,
            netcdf_version=netcdf_version,
            save_thresholds=save_thresholds,
            date_event=date_event,
            min_spell_length=min_spell_length,
            rolling_window_width=rolling_window_width,
            sampling_method=sampling_method,
        )
    if index_name is not None:
        return _build_standard_index_config(
            in_files=in_files,
            index_name=index_name,
            var_name=var_name,
            slice_mode=slice_mode,
            time_range=time_range,
            threshold=threshold,
            callback=callback,
            base_period_time_range=base_period_time_range,
            doy_window_width=doy_window_width,
            only_leap_years=only_leap_years,
            ignore_Feb29th=ignore_Feb29th,
            interpolation=interpolation,
            out_unit=out_unit,
            netcdf_version=netcdf_version,
            save_thresholds=save_thresholds,
            date_event=date_event,
            min_spell_length=min_spell_length,
            rolling_window_width=rolling_window_width,
            sampling_method=sampling_method,
        )
    msg = "You must fill either index_name or user_index" "to compute a climate index."
    raise InvalidIcclimArgumentError(msg)


def _get_reference_period(
    base_period_time_range: Sequence[dt.datetime | str] | None,
) -> tuple | None:
    if base_period_time_range is not None:
        return tuple(
            (read_date(t).strftime("%m-%d-%Y") for t in base_period_time_range),
        )
    return None


def _parse_threshold(
    threshold: str | Threshold | Sequence[str | Threshold] | None,
    doy_window_width: int,
    reference_period: Sequence[dt.datetime | str] | None,
    only_leap_years: bool,
    interpolation: QuantileInterpolation,
) -> Threshold | Sequence[Threshold] | None:
    if isinstance(threshold, Threshold):
        return threshold
    if isinstance(threshold, str):
        return _build_threshold(
            threshold,
            doy_window_width=doy_window_width,
            reference_period=reference_period,
            only_leap_years=only_leap_years,
            interpolation=interpolation,
        )
    if isinstance(threshold, Sequence):
        return [
            _build_threshold(
                t,
                doy_window_width=doy_window_width,
                reference_period=reference_period,
                only_leap_years=only_leap_years,
                interpolation=interpolation,
            )
            for t in threshold
        ]
    return None


def _build_user_index_config(
    user_index: UserIndexDict,
    in_files: InFileLike,
    index_name: str | None,
    var_name: str | Sequence[str] | None,
    slice_mode: FrequencyLike | Frequency,
    time_range: Sequence[dt.datetime | str] | None,
    callback: Callable[[int], None],
    base_period_time_range: Sequence[dt.datetime] | Sequence[str] | None,
    doy_window_width: int,
    only_leap_years: bool,
    ignore_Feb29th: bool,  # noqa: N803
    interpolation: str | QuantileInterpolation,
    out_unit: str | None,
    netcdf_version: str | NetcdfVersion,
    save_thresholds: bool,
    date_event: bool,
    min_spell_length: int | None,
    rolling_window_width: int | None,
    sampling_method: SamplingMethodLike,
) -> IndexConfig:
    interpolation = QuantileInterpolationRegistry.lookup(interpolation)
    indicator = parse.read_indicator(user_index)
    sampling_frequency = FrequencyRegistry.lookup(slice_mode)
    threshold = parse.read_thresholds(
        user_index,
        doy_window_width=doy_window_width,
        reference_period=base_period_time_range,
        only_leap_years=only_leap_years,
        interpolation=interpolation,
    )
    logical_link = parse.read_logical_link(user_index)
    coef = parse.read_coef(user_index)
    date_event = parse.read_date_event(user_index)
    rename = index_name or user_index.get("index_name", None) or "user_index"
    output_unit = out_unit
    rolling_window_width = user_index.get("window_width", rolling_window_width)
    reference_period = _get_reference_period(
        user_index.get("ref_time_range", base_period_time_range)
    )
    climate_vars_dict = build_input_dict(
        in_files=in_files,
        var_names=var_name,
        threshold=threshold,
        standard_index=None,
    )
    is_compared_to_ref = _must_add_reference_var(climate_vars_dict, reference_period)
    climate_vars = build_climate_vars(
        climate_vars_dict=climate_vars_dict,
        ignore_Feb29th=ignore_Feb29th,
        time_range=time_range,
        base_period=reference_period,
        standard_index=None,
        is_compared_to_reference=is_compared_to_ref,
    )
    return IndexConfig(
        save_thresholds=save_thresholds,
        frequency=sampling_frequency,
        climate_variables=climate_vars,
        min_spell_length=min_spell_length,
        rolling_window_width=rolling_window_width,
        out_unit=output_unit,
        netcdf_version=NetcdfVersionRegistry.lookup(netcdf_version),
        interpolation=interpolation,
        callback=callback,
        is_compared_to_reference=is_compared_to_ref,
        reference_period=reference_period,
        indicator_name=indicator.name,
        logical_link=logical_link,
        coef=coef,
        date_event=date_event,
        sampling_method=sampling_method,
        rename=rename,
        indicator=indicator,
        reference=ICCLIM_REFERENCE,
    )


def _build_standard_index_config(
    in_files: InFileLike,
    index_name: str | GenericIndicator | StandardIndex,
    var_name: str | Sequence[str] | None,
    slice_mode: FrequencyLike | Frequency,
    time_range: Sequence[dt.datetime | str] | None,
    threshold: str | Threshold | Sequence[str | Threshold] | None,
    callback: Callable[[int], None],
    base_period_time_range: Sequence[dt.datetime] | Sequence[str] | None,
    doy_window_width: int,
    only_leap_years: bool,
    ignore_Feb29th: bool,  # noqa: N803
    interpolation: str | QuantileInterpolation,
    out_unit: str | None,
    netcdf_version: str | NetcdfVersion,
    save_thresholds: bool,
    date_event: bool,
    min_spell_length: int | None,
    rolling_window_width: int | None,
    sampling_method: SamplingMethodLike,
) -> IndexConfig:
    interpolation = QuantileInterpolationRegistry.lookup(interpolation)
    # logical link here link two climate_variable computations as with user_index.
    # It isalways AND for standard indices.
    logical_link = LogicalLinkRegistry.LOGICAL_AND
    sampling_frequency = FrequencyRegistry.lookup(slice_mode)
    coef = None
    index = _parse_index_kind(index_name)
    if isinstance(index, StandardIndex):
        standard_index = index.clone()
        indicator = standard_index.indicator.clone()
        threshold = standard_index.threshold
        rename = standard_index.short_name
        output_unit = out_unit or standard_index.output_unit
        reference = standard_index.reference
        indicator_name = standard_index.short_name
    elif isinstance(index, GenericIndicator):
        rename = None
        output_unit = out_unit
        standard_index = None
        indicator = index.clone()
        reference = ICCLIM_REFERENCE
        indicator_name = indicator.name
    else:
        err = f"Unknown index_name : `{index_name}`"
        raise InvalidIcclimArgumentError(err)
    reference_period = _get_reference_period(base_period_time_range)
    threshold = _parse_threshold(
        threshold,
        doy_window_width=doy_window_width,
        reference_period=reference_period,
        only_leap_years=only_leap_years,
        interpolation=interpolation,
    )
    climate_vars_dict = build_input_dict(
        in_files=in_files,
        var_names=var_name,
        threshold=threshold,
        standard_index=standard_index,
    )
    is_compared_to_ref = _must_add_reference_var(climate_vars_dict, reference_period)
    climate_vars = build_climate_vars(
        climate_vars_dict=climate_vars_dict,
        ignore_Feb29th=ignore_Feb29th,
        time_range=time_range,
        base_period=reference_period,
        standard_index=standard_index,
        is_compared_to_reference=is_compared_to_ref,
    )
    return IndexConfig(
        save_thresholds=save_thresholds,
        frequency=sampling_frequency,
        climate_variables=climate_vars,
        min_spell_length=min_spell_length,
        rolling_window_width=rolling_window_width,
        out_unit=output_unit,
        netcdf_version=NetcdfVersionRegistry.lookup(netcdf_version),
        interpolation=interpolation,
        callback=callback,
        is_compared_to_reference=is_compared_to_ref,
        reference_period=reference_period,
        indicator_name=indicator_name,
        logical_link=logical_link,
        coef=coef,
        date_event=date_event,
        sampling_method=sampling_method,
        rename=rename,
        indicator=indicator,
        reference=reference,
    )


def _parse_index_kind(
    index_name: StandardIndex | GenericIndicator | str,
) -> StandardIndex | GenericIndicator:
    if isinstance(index_name, str):
        index = EcadIndexRegistry.lookup_no_error(index_name)
        if index is None:
            index = DcscIndexRegistry.lookup_no_error(index_name)
        if index is None:
            index = GenericIndicatorRegistry.lookup(index_name)
        return index
    return index_name


def _write_output_file(
    result_ds: xr.Dataset,
    input_time_encoding: dict | None,
    netcdf_version: NetcdfVersion,
    file_path: str,
) -> None:
    """Write `result_ds` to a netCDF file on `out_file` path."""
    if input_time_encoding:
        time_encoding = {
            "calendar": input_time_encoding.get("calendar"),
            UNITS_KEY: input_time_encoding.get(UNITS_KEY),
            # Force float64 to avoid serialization warning
            "dtype": np.float64,
        }
    else:
        time_encoding = {
            UNITS_KEY: "days since 1850-1-1",
            "dtype": np.float64,  # force float
        }
    result_ds.to_netcdf(
        file_path,
        format=netcdf_version.name,
        encoding={"time": time_encoding},
    )


def _handle_deprecated_params(
    index_name: str | GenericIndicator | StandardIndex | None,
    user_index: UserIndexDict | None,
    save_thresholds: bool,
    indice_name: str | None,
    transfer_limit_Mbytes: float | None,  # noqa: N803
    user_indice: UserIndexDict | None,
    save_percentile: bool | None,
    window_width: int | None,
    doy_window_width: int | None,
) -> tuple[str, UserIndexDict, bool, int]:
    if indice_name is not None:
        log.deprecation_warning(old="indice_name", new="index_name")
        index_name = indice_name
    if user_indice is not None:
        log.deprecation_warning(old="user_indice", new="user_index")
        user_index = user_indice
    if transfer_limit_Mbytes is not None:
        log.deprecation_warning(old="transfer_limit_Mbytes")
    if save_percentile is not None:
        log.deprecation_warning(old="save_percentile", new="save_thresholds")
        save_thresholds = save_percentile
    if window_width is not None:
        log.deprecation_warning(old="window_width", new="doy_window_width")
        doy_window_width = window_width
    return index_name, user_index, save_thresholds, doy_window_width


def _setup(
    callback: Callable[[int], None],
    callback_start_value: int,
    logs_verbosity: Verbosity | str,
) -> None:
    # make xclim input daily check a warning instead of an error
    # TODO @bzah: it might be safer to feed a context manager which will setup
    #             and teardown these confs
    # https://github.com/cerfacs-globc/icclim/issues/289
    xclim.set_options(data_validation="warn")
    # keep attributes through xarray operations
    xr.set_options(keep_attrs=True)
    log.set_verbosity(logs_verbosity)
    log.start_message()
    callback(callback_start_value)


def _get_unit(output_unit: str | None, da: DataArray) -> str | None:
    da_unit = da.attrs.get(UNITS_KEY, None)
    if da_unit is None:
        if output_unit is None:
            warn(
                "No unit computed or provided for the index was found."
                " Use out_unit parameter to add one.",
                stacklevel=2,
            )
            return ""
        return output_unit
    return da_unit


def _compute_climate_index(
    climate_index: Indicator,
    config: IndexConfig,
    initial_history: str | None,
    initial_source: str | None,
    reference: str,
    rename: str | None = None,
) -> Dataset:
    result_da = climate_index(config)
    if rename:
        result_da = result_da.rename(rename)
    else:
        result_da = result_da.rename(climate_index.name)
    result_da.attrs[UNITS_KEY] = _get_unit(config.out_unit, result_da)
    if (
        config.frequency.post_processing is not None
        and "time" in result_da.dims
        and not isinstance(
            climate_index,
            (
                StandardizedPrecipitationIndex6,
                StandardizedPrecipitationIndex3,
            ),
        )
    ):
        resampled_da, time_bounds = config.frequency.post_processing(result_da)
        result_ds = resampled_da.to_dataset()
        if time_bounds is not None:
            result_ds["time_bounds"] = time_bounds
            result_ds.time.attrs["bounds"] = "time_bounds"
    else:
        result_ds = result_da.to_dataset()
    if config.save_thresholds:
        result_ds = xr.merge(
            [result_ds, _format_thresholds_for_export(config.climate_variables)],
        )
    history = _build_history(result_da, config, initial_history, climate_index)
    return _add_ecad_index_metadata(
        result_ds,
        climate_index,
        history,
        initial_source,
        reference,
    )


def _add_ecad_index_metadata(
    result_ds: Dataset,
    computed_index: Indicator,
    history: str,
    initial_source: str | None,
    reference: str,
) -> Dataset:
    result_ds.attrs.update(
        {
            "title": computed_index.standard_name,
            "references": reference,
            "institution": "Climate impact portal (https://climate4impact.eu)",
            "history": history,
            "source": initial_source if initial_source is not None else "",
            "Conventions": "CF-1.6",
        },
    )
    try:
        result_ds.lat.encoding["_FillValue"] = None
        result_ds.lon.encoding["_FillValue"] = None
    except AttributeError:
        try:
            result_ds.latitude.encoding["_FillValue"] = None
            result_ds.longitude.encoding["_FillValue"] = None
        except AttributeError:
            pass
    return result_ds


def _build_history(
    result_da: DataArray,
    config: IndexConfig,
    initial_history: str | None,
    indice_computed: Indicator,
) -> str:
    from icclim import __version__ as icclim_version

    if initial_history is None:
        # get xclim history
        initial_history = result_da.attrs[HISTORY_CF_KEY]
    else:
        # append xclim history
        initial_history = f"{initial_history}\n{result_da.attrs[HISTORY_CF_KEY]}"
    del result_da.attrs[HISTORY_CF_KEY]
    current_time = dt.datetime.now(tz=dt.timezone.utc).strftime(
        "%Y-%m-%d %H:%M:%S",
    )
    return (
        f"{initial_history}\n"
        f" [{current_time}]"
        f" Calculation of {indice_computed.name}"
        f" index ({config.frequency.adjective})"
        f" - icclim version: {icclim_version}"
    )


def _build_threshold(
    threshold: str | Threshold,
    doy_window_width: int,
    reference_period: Sequence[dt.datetime | str] | None,
    only_leap_years: bool,
    interpolation: QuantileInterpolation,
) -> Threshold:
    if isinstance(threshold, Threshold):
        return threshold
    return build_threshold(
        threshold,
        doy_window_width=doy_window_width,
        reference_period=reference_period,
        only_leap_years=only_leap_years,
        interpolation=interpolation,
    )


def _format_thresholds_for_export(climate_vars: list[ClimateVariable]) -> Dataset:
    return xr.merge([_format_threshold(v) for v in climate_vars])


def _format_threshold(cf_var: ClimateVariable) -> DataArray:
    return cf_var.threshold.value.rename(cf_var.name + "_thresholds").reindex()


def _must_add_reference_var(
    climate_vars_dict: dict[str, InFileDictionary],
    reference_period: Sequence[str] | None,
) -> bool:
    """
    Check if the reference variable must be added to the input variables.

    Return True whenever the input has no threshold and only one studied variable but
    there is a reference period.
    Example case: the anomaly of tx(1960-2100) by tx(1960-1990).
    """
    t = next(iter(climate_vars_dict.values())).get("thresholds", None)
    return t is None and len(climate_vars_dict) == 1 and reference_period is not None



"""Typing for generic indices."""

from __future__ import annotations

import abc
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from xarray import DataArray
    from xclim.core.missing import MissingBase


class MissingMethodLike(ABC):
    """Workaround xclim missing type."""

    # TODO @bzah: PR that to xclim
    # https://github.com/cerfacs-globc/icclim/issues/289

    @abstractmethod
    def execute(self, *args, **kwargs) -> MissingBase:
        """Execute the missing method."""
        ...

    @abstractmethod
    def validate(self, *args, **kwargs) -> bool:
        """Validate the missing method."""
        ...


class Indicator(ABC):
    """
    Generic indicator abstract class.

    Attributes
    ----------
    name: str
        The name of the indicator.
    standard_name: str
        The standard name of the indicator, ideally from the CF conventions.
    long_name: str
        The long name of the indicator.
    cell_methods: str
        The cell methods of the indicator.
    qualifiers: tuple
        The qualifiers of the indicator, used to classify indicators.
    templated_properties: tuple
        The properties that can be templated.
        Theses properties are used to fill the output metadata.
    """

    name: str
    standard_name: str
    long_name: str
    cell_methods: str
    qualifiers: tuple

    templated_properties = (
        "standard_name",
        "long_name",
        "cell_methods",
    )

    @abc.abstractmethod
    def __call__(self, *args, **kwargs) -> DataArray:
        """Compute the indicator."""
        ...

    @abc.abstractmethod
    def preprocess(self, *args, **kwargs) -> list[DataArray]:
        """Preprocess the data."""
        ...

    @abc.abstractmethod
    def postprocess(self, *args, **kwargs) -> DataArray:
        """Postprocess the data."""
        ...

    @abc.abstractmethod
    def __eq__(self, __value: object) -> bool:
        """Check if two indicators are equal."""
        ...

    def clone(self) -> Indicator:
        """Clone the indicator."""
        from copy import deepcopy

        return deepcopy(self)
